// main.cc
// -*-C++-*-
// $Id: main.cc,v 1.1 2004/06/13 12:43:09 steve Exp steve $
//
// Copyright (C) 2001 Steve Atkins <steve@blighty.com>
// Last change: Time-stamp: <13-Jun-2004 06:50:58 steve>
//
//

#include <qfile.h>
#include <iostream>
#include <qstring.h>
#include <qmap.h>
#include <qstringlist.h>
#include <qmemarray.h>


struct glyph {
  int c;
  int width;     // millipoints
  int kernright; // millipoints
};

struct chardata_t {
  int wx;
  QMap<QString,int> pairkern;
};

class PFont
{
public:
  PFont() {}
  PFont(const QStringList&, int idx);
  virtual ~PFont();

  virtual const char * isA() const { return "PFont"; };

  void load(const QStringList&);
  void dump();
  void dumparray();

  QMemArray<glyph> glyphWidths(const QString& s, int fontsize);

  int BoxLeft(int fs) const { return (boxl * fs) / 1000 ; }
  int BoxRight(int fs) const { return (boxr * fs) / 1000; }
  int BoxTop(int fs) const { return (boxt * fs) / 1000; }
  int BoxBottom(int fs) const { return (boxb * fs) / 1000; }

private:
  void ParseCharMetrics(QStringList::const_iterator& cit, const QStringList::const_iterator& end, int count);
  void ParseKernData(QStringList::const_iterator& cit, const QStringList::const_iterator& end);

  QMap<int,QString> charname;
  QMap<QString,chardata_t> chardata;
  QString fontname;
  QString fullname;
  QString familyname;
  int underlineposition;
  int underlinethickness;
  int capheight;
  int xheight;
  int ascender;
  int descender;
  int boxl;
  int boxr;
  int boxt;
  int boxb;
};

PFont::~PFont()
{
}

void PFont::dump()
{
  QString tag(fontname);
  tag.replace('-', "_");
  cout <<
    "  {\n"
    "    \"" << fontname << "\", // fontname\n"
    "    \"" << fullname << "\", // fullname\n"
    "    \"" << familyname << "\", //familyname\n"
    "    " << underlineposition << ", // underlineposition\n"
    "    " << underlinethickness << ", // underlinethickness\n"
    "    " << capheight << ", // capheight\n"
    "    " << xheight << ", // xheight\n"
    "    " << ascender << ", // ascender\n"
    "    " << descender << ", // descender\n"
    "    " << boxl << ", // boxl\n"
    "    " << boxr << ", // boxr\n"
    "    " << boxt << ", // boxt\n"
    "    " << boxb << ", // boxb\n"
    "    width_" << tag << ", // charwidth\n"
    "    kern_" << tag << " // pairwise kerning\n  },\n";
}

void PFont::dumparray()
{
  QString tag(fontname);
  tag.replace('-', "_");
  cout << "const charwidthData width_" << tag << "[] = { // glyph widths\n";
  for(QMap<int,QString>::const_iterator cit=charname.begin(); cit!=charname.end(); ++cit) {
    cout <<
      "  {" << cit.key() << ", " << chardata[cit.data()].wx << "}, // " << cit.data() << "\n";
  }
  cout <<
    "  {0,0}\n};\n"
    "const kernData kern_" << tag << "[] = { // Pairwise kerning\n";
  for(QMap<int,QString>::const_iterator cit=charname.begin(); cit!=charname.end(); ++cit) {
    for(QMap<QString,int>::const_iterator kit=chardata[cit.data()].pairkern.begin(); kit!=chardata[cit.data()].pairkern.end(); ++kit) {
      for(QMap<int,QString>::const_iterator cit2=charname.begin(); cit2!=charname.end(); ++cit2) {
	if(cit2.data() == kit.key()) {
	  cout <<
	    "  {" << cit.key() << ", " << cit2.key() <<
	    ", " << kit.data() << "}, // " <<
	    cit.data() << " " << cit2.data() << "\n";
	}
      }
    }
  }
  cout <<
    "  {0,0,0}\n"
    "};\n";
}

void PFont::load(const QStringList& in)
{
  QStringList::const_iterator cit=in.begin();
  while(cit != in.end()) {
    QString entry((*cit).section(' ', 0, 0));
    QString remainder((*cit).section(' ', 1));

    if(entry == "FontName") {
      fontname = remainder;
    } else if(entry == "FullName") {
      fullname = remainder;
    } else if(entry == "FamilyName") {
      familyname = remainder;
    } else if(entry == "UnderlinePosition") {
      underlineposition = remainder.toInt();
    } else if(entry == "UnderlineThickness") {
      underlinethickness = remainder.toInt();
    } else if(entry == "CapHeight") {
      capheight = remainder.toInt();
    } else if(entry == "XHeight") {
      xheight = remainder.toInt();
    } else if(entry == "Ascender") {
      ascender = remainder.toInt();
    } else if(entry == "Descender") {
      descender = remainder.toInt();
    } else if(entry == "FontBBox") {
      boxl = remainder.section(' ', 0, 0).toInt();
      boxb = remainder.section(' ', 1, 1).toInt();
      boxr = remainder.section(' ', 2, 2).toInt();
      boxt = remainder.section(' ', 3, 3).toInt();
    } else if(entry == "StartCharMetrics") {
      ParseCharMetrics(cit, in.end(), remainder.toInt());
    } else if(entry == "StartKernData") {
      ParseKernData(cit, in.end());
    }

    if(cit!=in.end()) {
      ++cit;
    }
  }
}

void PFont::ParseCharMetrics(QStringList::const_iterator& cit, const QStringList::
const_iterator& end, int count)
{
  ++cit;
  while(cit != end && count--) {
    QStringList line = QStringList::split(';', *cit);
    QString name;
    int wx=-1;
    int c;
    for(QStringList::const_iterator lit=line.begin(); lit!=line.end(); ++lit) {
      QString t((*lit).section(' ', 0, 0, QString::SectionSkipEmpty));
      QString v((*lit).section(' ', 1, 0xfffffff, QString::SectionSkipEmpty));
      if(t == "C") {
        c = v.toInt();
      } else if(t == "WX") {
        wx = v.toInt();
      } else if(t == "N") {
        name = v;
      }
    }
    if(name.isNull()) {
      cerr << "Expecting character name in " << *cit << "\n";
    } else {
      chardata[name].wx = wx;
      if(c >= 0) {
        charname[c] = name;
      }
    }
    ++cit;
  }
  if(*cit != "EndCharMetrics") {
    cerr << "Expected EndCharMetrics, got " << *cit << "\n";
  }
}

void PFont::ParseKernData(QStringList::const_iterator& cit, const QStringList::const_iterator& end)
{
  ++cit;
  while(*cit != "EndKernData") {
    if(cit == end)
      return;
    if((*cit).section(' ', 0, 0) == "StartKernPairs") {
      int count = (*cit).section(' ', 1, 1).toInt();
      while(count--) {
        if(*cit == "EndKernPairs")
          break;
        if(*cit == "EndKernData" || cit == end)
          return;

        QStringList line=QStringList::split(' ', *cit);
        if(line[0] == "KPX") {
          chardata[line[1]].pairkern[line[2]] = line[3].toInt();
        }
        ++cit;
      }
    }
    ++cit;
  }
}

QMemArray<glyph> PFont::glyphWidths(const QString& s, int fontsize)
{
  QMemArray<glyph> ret(s.length());
  
  for(int i=0; i<(int)s.length(); ++i) {
    int c=s[i].unicode();
    ret[i].c=c;
    QMap<int,QString>::const_iterator cit=charname.find(c);
    if(cit == charname.end()) {
      cerr << "Unrecognised glyph " << c << "\n";
      ret[i].width = fontsize*1000;
      ret[i].kernright = 0;
    } else {
      ret[i].width = chardata[cit.data()].wx * fontsize;
      ret[i].kernright = 0;
      if(i < (int)s.length() -1) {
        // Kerning
        QMap<int,QString>::const_iterator cit2 = charname.find(s[i+1].unicode());
        if(cit2 != charname.end()) {
          ret[i].kernright = chardata[cit.data()].pairkern[cit2.data()] * fontsize
;
        }
      }
    }
  }
  return ret;
}


int main(int argc, char **argv)
{
  if(argc < 2) {
    cerr << "Usage: fontdata <filename>\n";
    exit(1);
  }
  PFont f[50];
  for(int i=1; i<argc; i++) {
    QString filename(argv[i]);
    QFile inf(filename);
    if(!inf.open(IO_ReadOnly)) {
      cerr << "Failed to open " << filename << "\n";
      exit(1);
    }
    
    QStringList afmtext;
    QString line;
    while(-1 != inf.readLine(line, 10000)) {
      afmtext.append(line.stripWhiteSpace());
    }
    inf.close();
    
    f[i].load(afmtext);
    f[i].dumparray();
  }

  cout << "const fontData staticfont[] = {\n";
  for(int i=1; i<argc; ++i) {
    f[i].dump();
  }
  cout << "  {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}\n};\n";

  return 0;
}
