// main.cc
// -*-C++-*-
// $Id: main.cc,v 1.1 2004/06/13 01:30:44 steve Exp steve $
//
// Copyright (C) 2001 Steve Atkins <steve@blighty.com>
// Last change: Time-stamp: <25-Jul-2004 20:08:48 steve>
//
//

#include "pdf.h"
#include "pdfdocument.h"
#include "pdfPage.h"
#include "pdfContent.h"
#include "qimage.h"
#include "pdfImage.h"
#include "pdfText.h"
#include "pdfGrid.h"

#include <qstring.h>
#include <qstringlist.h>
#include <qfile.h>
#include <qdom.h>
#include <qsqldatabase.h>
#include <qsqlrecord.h>
#include <qapplication.h>
#include <qvaluevector.h>
#include <unistd.h>
#include <getopt.h>

#ifndef NOCHART
#include "chartdir.h"
#endif

int globaldebug=0;

int chartspin=0;

QSqlDatabase *db;

QMap<QString,pdf::hAlign> halign;

void processfile(QString fn);
void initdata();

QMap<QString,QString> globalvar;

int rowcount=0;
bool writepdf = true;
bool writexml = false;

int main(int argc, char **argv)
{
#ifndef NOCHART
  setLicenseCode("your-license-code-here");
#endif
  QApplication app(argc, argv, false);

  QString dbhost("127.0.0.1");
  QString dbport("5432");
  QString dbuser("abacus");
  QString dbpass("abacus");
  QString dbdb("abacus");
  QString gvar;

  initdata();
  QStringList inputfiles;
  while(1) {
    int option_index = 0;
    static struct option long_options[] = {
      {"database", required_argument, 0, 'd'},
      {"host",     required_argument, 0, 'h'},
      {"port",     required_argument, 0, 'p'},
      {"user",     required_argument, 0, 'U'},
      {"pass",     required_argument, 0, 'P'},
      {"define",   required_argument, 0, 'D'},
      {"xml",      no_argument,       0, 'x'},
      {"debug",    required_argument, 0, 'Z'},
      {0,0,0,0}
    };
    int c = getopt_long(argc, argv, "d:h:p:U:P:D:xZ", long_options, &option_index);
    if(-1==c)
      break;
    switch(c) {
    case 'd':
      dbdb = optarg;
      break;
    case 'h':
      dbhost = optarg;
      break;
    case 'p':
      dbport = optarg;
      break;
    case 'U':
      dbuser = optarg;
      break;
    case 'P':
      dbpass = optarg;
      break;
    case 'D':
      gvar = optarg;
      globalvar[gvar.section("=", 0, 0)] = gvar.section("=", 1);
      break;
    case 'x':
      writepdf = false;
      writexml = true;
      break;
    case 'Z':
      globaldebug = atoi(optarg);
    case '?':
      break;
    default:
      break;
    }
  }
  while(optind < argc) {
    inputfiles.append(argv[optind++]);
  }
  db = QSqlDatabase::addDatabase("QPSQL7");
  //  QString dbstring="dbname=" + dbdb + ";host=" + dbhost + ";port=" + dbport;
  db->setHostName(dbhost);
  db->setPort(dbport.toInt());
  db->setDatabaseName(dbdb);
  db->setUserName(dbuser);
  db->setPassword(dbpass);

  if(!db->open()) {
    cerr << "Failed to open database: " << db->lastError().driverText() <<
      " " << db->lastError().databaseText() << "\n";
    exit(1);
  }

  if(inputfiles.empty()) {
    cerr << "No input files specified\n";
    exit(0);
  }

  QSqlQuery series("CREATE TEMPORARY TABLE series (n integer not null)", db);
  if(!series.isActive()) {
    cerr << "Failed to create series table\n";
    exit(1);
  }
  
  QSqlQuery insert(db);
  insert.prepare("INSERT INTO series (n) VALUES (:c)");
  for(int i=0; i<365; i++) {
    insert.bindValue(":c", i);
    insert.exec();
  }

  for(QStringList::const_iterator cit=inputfiles.begin(); cit!=inputfiles.end(); ++cit) {
    processfile(*cit);
  }
  return 0;
}

void walkdoc(const QDomElement &parent, int depth);

void walkdoc(const QDomElement &parent, int depth)
{
  QDomNode node = parent.firstChild();
  while(!node.isNull()) {
    if(node.isElement()) {
      QDomElement element = node.toElement();
      cerr << "el ";
      for(int i=0; i<depth; i++) {
	cerr << " ";
      }
      cerr << element.tagName() << "\n";
      QDomNamedNodeMap atts = element.attributes();
      for(unsigned int j=0; j<atts.length(); ++j) {
	QDomAttr a = atts.item(j).toAttr();
	if(!a.isNull()) {
	  for(int i=0; i<depth; i++) {
	    cerr << " ";
	  }
	  cerr << "     " << a.name() << " = " << a.value() << "\n";
	}
      }
      walkdoc(element, depth+1);
    }
    if(node.isText()) {
      QDomText text = node.toText();
      cerr << "tx ";
      for(int i=0; i<depth; i++) {
	cerr << " ";
      }
      cerr << text.data() << "\n";
    }
    node = node.nextSibling();
  }
}

void inheritattribute(const QString& attribute, const QDomElement& parent, QDomElement& child)
{
  if(parent.hasAttribute(attribute) && !child.hasAttribute(attribute))
    child.setAttribute(attribute, parent.attribute(attribute));
}

pdf::hAlign gethalign(QString s)
{
  QMap<QString,pdf::hAlign>::const_iterator cit = halign.find(s);
  if(cit == halign.end()) {
    cerr << "Bad alignment: '" << s << "'\n";
    exit(1);
  }
  return cit.data();
}

void doFontTags(QDomElement &fonttag, pdfText& t)
{
  QString font = fonttag.attribute("font");
  if(!font.isNull()) {
    t.setFont(font.latin1());
  }
  QString pointsize = fonttag.attribute("size");
  if(!pointsize.isNull()) {
    bool ok=false;
    double ps = pointsize.toDouble(&ok);
    if(!ok || ps < 0.0) {
      cerr << "Invalid font size: '" << pointsize << "'\n";
      exit(1);
    }
    t.setPointsize(ps);
  }
  t.setHAlign(gethalign(fonttag.attribute("align", "left")));
}

#ifndef NOCHART

int chartColor(BaseChart *c, QString color, double angle=90.0, double scale=1.0);

int chartColor(BaseChart *c, QString color, double angle, double scale)
{
  //  cerr << "chartColor(" << color << ")\n";
  if(color == "Transparent") {
    return Transparent;
  }
  if(color=="goldGradient") {
    return c->gradientColor(goldGradient);
  }
  if(color=="silverGradient") {
    return c->gradientColor(silverGradient);
  } 
  if(color=="redGradient") {
    return c->gradientColor(redMetalGradient);
  }
  if(color=="greenGradient") {
    return c->gradientColor(greenMetalGradient);
  }
  if(color=="blueGradient") {
    return c->gradientColor(blueMetalGradient);
  }

  int nsep = color.contains(':');
  if(nsep == 0) {
    // It's a flat colour
    unsigned int al=0;
    if(color.contains('%')) {
      // Handle the alpha channel
      bool ok;
      double alpha = color.section("%", 0, 0).toDouble(&ok);
      if(!ok || alpha < 0.0 || alpha > 100.0) {
	cerr << "Bad alpha parameter in '" << color << "'\n";
	exit(1);
      }
      al = 255 - (int)(alpha * 2.55);
      //      cerr << "alpha=" << alpha << " al=" << al << endl;
      if(al==255) {
	return Transparent;
      }
      color = color.section("%", 1);
    }
    QColor col(color);
    if(!col.isValid()) {
      cerr << "Bad color specification in '" << color << "'\n";
      exit(1);
    }
    //    cerr << "col.rgb=" << hex << col.rgb() << endl;
    unsigned int ret=(al<<24)|(col.rgb() & 0xffffff);
    //    cerr << "al=" << al << " Returning " << hex << ret << endl;
    return ret;
  }

  const int maxcolors=100;
  int ncolors=0;
  int colorlist[maxcolors*2+2];
  QStringList colors = QStringList::split(':', color);
  int np = colors.count();
  if(np > maxcolors) {
    cerr << "Absurdly long colour list truncated\n";
    np = maxcolors;
  }
  int x=0;
  int dx = 256 / np;
  QStringList::const_iterator cit=colors.begin();
  while(ncolors < maxcolors && cit != colors.end()) {
    QString thiscolor = *cit++;
    int delta=dx;
    if(thiscolor.contains('.')) {
      bool ok;
      delta = thiscolor.section('.', 1, 1).toInt(&ok);
      if(!ok || delta < 0 || delta+x > 256) {
	cerr << "Bad color delta '" << thiscolor << "' in " << color << "\n";
	exit(1);
      }
      thiscolor = thiscolor.section('.', 0, 0);
    }
    colorlist[ncolors++] = x;
    colorlist[ncolors++] = chartColor(c, thiscolor);
    x+=delta;
  }
  return c->gradientColor(IntArray(colorlist, ncolors), angle, scale, 0, 0);
}

QString chartText(QDomElement &parent)
{
  /*
  QString ret;
  QDomNode node=parent.firstChild();
  while(!node.isNull()) {
    if(node.isText()) {
      QDomText text=node.toText();
      ret += text.data();
    }
  }
  return ret;
  */
  return parent.text();
}

void walkChart(QDomElement &parent, pdfText& t)
{
  //  cerr << "Walking chart\n";
  double width=parent.attribute("width", "0.0").toDouble();
  double height=parent.attribute("height", "0.0").toDouble();
  double pixel=parent.attribute("pixel", "1.0").toDouble();
  QString bgcolor=parent.attribute("bg", "white");
  double bgangle=parent.attribute("bgangle", "90.0").toDouble();
  double bgscale=parent.attribute("bgscale", "1.0").toDouble();
  QString bordercolor=parent.attribute("bordercolor", "Transparent");
  int borderwidth=parent.attribute("borderwidth", "1.0").toInt();
  int barborderwidth=parent.attribute("barborderwidth", "0").toInt();
  double pad= parent.attribute("pad", "30.0").toDouble();
  QString type=parent.attribute("type");
  double legendwidth=parent.attribute("legendwidth", "0.0").toInt();

  if(width==0.0 && height==0.0) {
    width = t.document()->usableWidth();
  }
  if(width==0.0) {
    width = height / 0.8;
  }
  if(height==0.0) {
    height = width * 0.8;
  }

  int pwidth = (int)(width * pixel);
  int pheight = (int)(height * pixel);
  int ppad = (int)(pad * pixel);
  int plegend = (int)(legendwidth * pixel);
  int toppad = ppad;
  int rightpad = ppad;
  int leftpad = ppad;
  int bottompad = ppad;
  if(plegend) {
    rightpad += 10 + plegend;
  }

  bgscale *= (pwidth / 256.0);

  XYChart *c = new XYChart(pwidth, pheight);

  c->setBackground(chartColor(c, bgcolor, bgangle, bgscale), chartColor(c, bordercolor), 0);

  QDomNode node = parent.firstChild();
  QMap<QString, QMap<QString,double> > data;
  QStringList groups;
  QStringList columnnames;
  while(!node.isNull()) {
    if(node.isElement()) {
      QDomElement element = node.toElement();
      if(element.tagName() == "title") {
	//	cerr << "Doing title\n";
	QString font=element.attribute("font", "bold");
	double fontSize=element.attribute("size", "12.0").toDouble();
	int fontColor=chartColor(c, element.attribute("color", "black"));
	int edgeColor=chartColor(c, element.attribute("bordercolor", "Transparent"));
	int bgColor=chartColor(c, element.attribute("bg", "Transparent"));
	QString text = chartText(element);
	TextBox *t=c->addTitle(text.utf8(), font.utf8(), fontSize, fontColor, bgColor, edgeColor);
	//	cerr << "Adding Title: '" << text.latin1() << "' in font '" << font.latin1() << "' size=" << fontSize << " color=" << hex << fontColor << endl;
	toppad += t->getHeight();
      }
      if(element.tagName() == "data") {
	//	cerr << "Doing data\n";
	QString defaultgroup = element.attribute("group");
	if(0==groups.contains(defaultgroup)) {
	  groups.append(defaultgroup);
	}
	int defaultcolor = chartColor(c, element.attribute("color", "black"));
	QDomNode node2=element.firstChild();
	while(!node2.isNull()) {
	  if(node2.isElement()) {
	    QDomElement el2 = node2.toElement();
	    if(el2.tagName() == "d") {
	      QString nc = el2.attribute("color");
	      int color=nc.isEmpty() ? defaultcolor : chartColor(c, nc);
	      QString group = el2.attribute("group", defaultgroup);
	      //	      cerr << "Group " << group << "\n";
	      if(0==groups.contains(group)) {
		groups.append(group);
	      }
	      QDomNode node3=el2.firstChild();
	      QString x;
	      double y=0.0;
	      while(!node3.isNull()) {
		if(node3.isElement()) {
		  QDomElement el3=node3.toElement();
		  if(el3.tagName() == "x") {
		    x = el3.text();
		    if(0==columnnames.contains(x)) {
		      columnnames.append(x);
		    }
		  } else if(el3.tagName() == "y") {
		    bool ok;
		    y = el3.text().toDouble(&ok);
		    if(!ok) {
		      cerr << "Bad <y> value: '" << el3.text() << "'\n";
		      exit(1);
		    }
		  }
		}
		node3 = node3.nextSibling();
	      }
	      if(x.isNull()) {
		cerr << "Missing <x> parameter\n";
		exit(1);
	      }

	      data[group][x] = y;
	    }
	  }
	  node2 = node2.nextSibling();
	}

      }
    }
    node = node.nextSibling();
  }
  //  cerr << "Finished reading chart\n";
  c->setPlotArea(leftpad, bottompad, pwidth-leftpad-rightpad, pheight-toppad-bottompad);

  if(plegend) {
    c->addLegend(pwidth - (plegend + rightpad) / 2, toppad);
  }
  
  const char **labels = new (const char*)[columnnames.count()];
  int ncol=0;
  QValueVector<QCString> lbls;
  for(QStringList::const_iterator cit=columnnames.begin(); cit!=columnnames.end(); ++cit) {
    //    cerr << "L: " << *cit << "\n";
    lbls.append((*cit).utf8());
    labels[ncol++] = lbls.back();
  }
  for(int i=0; i<ncol; i++) {
    //    cerr << "Label: " << labels[i] << endl;
  }

  //  const char *l2[] = {"foo", "bar", "baz"};

  //  c->xAxis()->setLabels(StringArray(l2, 3));
  c->xAxis()->setLabels(StringArray(labels, ncol))->setFontAngle(90);

  double *ds = new double[columnnames.count()];

  //  cerr << "Creating layer\n";
  BarLayer *barlayer = c->addBarLayer(Stack, barborderwidth);
  barlayer->setBorderColor(-1, 0);
  for(QStringList::const_iterator git=groups.begin(); git!=groups.end(); ++git) {
    for(int i=0; i<(int)columnnames.count(); ++i) {
      ds[i] = 0.0;
    }
    int i=0;
    for(QStringList::const_iterator cit=columnnames.begin(); cit!=columnnames.end(); ++cit) {
      ds[i++] = data[*git][*cit];
    }
    barlayer->addDataSet(DoubleArray(ds, columnnames.count()), -1, (*git).latin1());
  }
  delete[] ds;

  //  cerr << "Making chart\n";
  //c->makeChart("foo.png");
  MemBlock imb = c->makeChart(0); // PNG
  //  cerr << "Chart made\n";
  QImage im;
  im.loadFromData((unsigned char*)(imb.data), imb.len, "PNG");

  delete[] labels;
  delete c;

  char spinbuff[50];
  sprintf(spinbuff, "chart_%d", chartspin++);

  t.append(spinbuff, im, width, height);
  //  cerr << "Done with chart\n";
}
#endif

void walkText(QDomElement &parent, pdfText& t)
{
  //  cerr << "Walking text\n";
  QDomNode node = parent.firstChild();
  while(!node.isNull()) {
    //    cerr << "Top of node loop\n";
    if(node.isElement()) {
      QDomElement element = node.toElement();
      //      cerr << "gotElement " << element.tagName() << "\n";
      if(element.tagName() == "b") {
	bool bold = t.bold();
	t.setBold(true);
	walkText(element, t);
	t.setBold(bold);
      } else if(element.tagName() == "i") {
	bool italic = t.italic();
	t.setItalic(true);
	walkText(element, t);
	t.setItalic(italic);
      } else if(element.tagName() == "u") {
	bool underline = t.underline();
	t.setUnderline(true);
	walkText(element, t);
	t.setUnderline(underline);
      } else if(element.tagName() == "font" || element.tagName() == "t") {
	QCString font = t.font();
	double pointsize = t.pointsize();
	doFontTags(element, t);
	walkText(element, t);
	t.setFont(font);
	t.setPointsize(pointsize);
      } else if(element.tagName() == "br") {
	QString pointsize = element.attribute("size", "0.0");
	bool ok = false;
	double ps = pointsize.toDouble(&ok);
	if(!ok || ps < 0.0) {
	  cerr << "Invalid size in <br/>: '" << pointsize << "'\n";
	  exit(1);
	}
	t.appendBreak(ps);
      } else if(element.tagName() == "img") {
	QString filename = element.attribute("src");
	QString width = element.attribute("width", "0.0");
	QString height = element.attribute("height", "0.0");
	double w = width.toDouble();
	double h = height.toDouble();

	QImage img(filename);
	if(img.isNull()) {
	  cerr << "Failed to load image file '" << filename << "'\n";
	  exit(1);
	}
	t.append(filename.latin1(), img, w, h);
#ifndef NOCHART
      } else if(element.tagName() == "chart") {
	walkChart(element, t);
#endif
      } else {
	cerr << "Unrecognised tag found inside <t></t>: '" << element.tagName() << "'\n";
	exit(1);
      }
    } else if(node.isText()) {
      //      cerr << "Text node\n";
      QDomText text = node.toText();
      //      cerr << text.data() << "\n";
      t.append(text.data().latin1());
      //      cerr << "Done\n";
    } else if(!node.isComment()) {
      cerr << "Non-text, non-element node: '" << node.nodeValue() << "'\n";
    }
    //    cerr << "Going to next sibling\n";
    node = node.nextSibling();
  }
  //  cerr << "Leaving walkText\n";
}

void doText(QDomElement &parent, pdfdocument *d)
{
  //  cerr << "doText()\n";
  pdf::hAlign align = gethalign(parent.attribute("align", "left"));
  pdfText t(d);
  doFontTags(parent, t);
  walkText(parent, t);
  //  cerr << "Appending text to document\n";
  d->text(&t, align);
  //  cerr << "Leaving doText\n";
}

void walkRow(QDomElement &parent, pdfGrid& g, int row, QMap<int,bool>& cellused, pdfdocument *d)
{
  int col=0;
  rowcount++;

  QDomNode node = parent.firstChild();
  while(!node.isNull()) {
    if(node.isElement()) {
      QDomElement element = node.toElement();
      if(element.tagName() == "td") {
	inheritattribute("bg", parent, element);
	inheritattribute("font", parent, element);
	inheritattribute("size", parent, element);
	inheritattribute("align", parent, element);
	QString rowspan = element.attribute("rowspan", "1");
	bool ok=false;
	int rs = rowspan.toInt(&ok);
	if(!ok || rs < 1) {
	  cerr << "Bad rowspan element '" << rowspan << "'\n";
	  exit(1);
	}
	QString colspan = element.attribute("colspan", "1");
	ok=false;
	int cs = colspan.toInt(&ok);
	if(!ok || cs < 1) {
	  cerr << "Bad colspan element '" << colspan << "'\n";
	  exit(1);
	}

	QString bgcolor = element.attribute("bg", "white");
	int nels = bgcolor.contains(':');
	if(nels++) {
	  bgcolor = bgcolor.section(":", rowcount % nels, rowcount % nels);
	}

	while(cellused.find((row<<16)|col) != cellused.end())
	  ++col;
	for(int i=0; i<cs; i++) {
	  for(int j=0; j<rs; j++) {
	    cellused[((j+row)<<16)|(i+col)] = true;
	  }
	}
	pdfText t(d);
	doFontTags(element, t);
	walkText(element, t);
	pdfCell cell(t);
	if(!bgcolor.isEmpty()) {
	  cell.setCellColor(bgcolor);
	}
	g.insertCell(cell, col, row, cs, rs);
      } else {
	cerr << "Unrecognised element '" << element.tagName() << "' found in <tr>\n";
	exit(1);
      }
    } else if(!node.isComment()) {
      cerr << "Found non-element: " << node.nodeName() << " '" << node.nodeValue() << "'\n";
      exit(1);
    }
    node = node.nextSibling();
  }
}

void walkGrid(QDomElement &parent, pdfGrid& g, pdfdocument *d)
{
  QMap<int,bool> cellused;
  int row = 0;

  QDomNode node = parent.firstChild();
  while(!node.isNull()) {
    if(node.isElement()) {
      QDomElement element = node.toElement();
      if(element.tagName() == "tr") {
	inheritattribute("bg", parent, element);
	inheritattribute("font", parent, element);
	inheritattribute("size", parent, element);
	inheritattribute("align", parent, element);
	walkRow(element, g, row++, cellused, d);
      } else {
	cerr << "Unrecognised element '" << element.tagName() << "' found in <table>\n";
	exit(1);
      }
    } else if(!node.isComment()) {
      cerr << "Found non-element: " << node.nodeName() << " '" << node.nodeValue() << "'\n";
      exit(1);
    }
    node = node.nextSibling();
  }
}

void doGrid(QDomElement &parent, pdfdocument *d)
{
  //  cerr << "doGrid()\n";
  pdf::hAlign align = gethalign(parent.attribute("align", "left"));
  pdfGrid g;
  walkGrid(parent, g, d);
  d->grid(&g, align);
}

// We expect a sequence of <table> or <t>
void gendoc(const QDomElement &parent, pdfdocument& d)
{
  //  cerr << "gendoc()\n";
  QDomNode node = parent.firstChild();
  while(!node.isNull()) {
    //    cerr << "Top of gendoc loop\n";
    if(node.isElement()) {
      QDomElement el(node.toElement());
      if(el.nodeName() == "table") {
	doGrid(el, &d);
      } else if(el.nodeName() == "t") {
	doText(el, &d);
      } else if(el.nodeName() != "meta") {
	cerr << "Unrecognised element: " << el.nodeName() << "\n";
	exit(1);
      }
    } else if(!node.isComment()) {
      cerr << "Found non-element: " << node.nodeName() << " '" << node.nodeValue() << "'\n";
      exit(1);
    }

    node = node.nextSibling();
  }
  //  cerr << "Leaving gendoc\n";
}

void replacevariables(QDomNode& parent, const QMap<QString,QString>& var)
{
  if(var.empty())
    return;
  
  //  cerr << "  Entering replacevariables\n";

  if(parent.isElement()) {
    QDomElement element = parent.toElement();
    //    cerr << "    Element " << element.tagName() << "\n";
    QDomNamedNodeMap atts = element.attributes();
    for(unsigned int j=0; j<atts.length(); ++j) {
      QDomAttr a = atts.item(j).toAttr();
      if(!a.isNull()) {
	for(QMap<QString,QString>::const_iterator cit=var.begin(); cit!=var.end(); ++cit) {
	  // We avoid a blind replace as we don't know how expensive
	  // setValue() is.
	  if(a.value().contains(cit.key())) {
	    a.setValue(a.value().replace(cit.key(), cit.data()));
	  }
	}
      }
    }

    QDomNode node = element.firstChild();
    while(!node.isNull()) {
      replacevariables(node, var);
      node = node.nextSibling();
    }  
  }

  if(parent.isText()) {
    QDomText text = parent.toText();
    //    cerr << "    Checking on text '" << text.data() << "'\n";
    for(QMap<QString,QString>::const_iterator cit=var.begin(); cit!=var.end(); ++cit) {
      if(text.data().contains(cit.key())) {
	text.setData(text.data().replace(cit.key(), cit.data()));
      }
    }
  }
}

void expandwalk(QDomElement& parent)
{
  QDomNode node = parent.firstChild();
  while(!node.isNull()) {
    QDomNode nextnode = node.nextSibling();
    if(node.isElement()) {
      QDomElement element(node.toElement());
      if(element.tagName() == "sql") {
	QString name = element.attribute("name");
	QString prefix = name.isEmpty() ? "$" : ("$" + name + ".");
	QString query = element.attribute("query");
	//	cerr << "Prepping '" << query << "'\n";
	QSqlQuery q(db);
	q.prepare(query);
	// bind attributes
	QDomNamedNodeMap atts = element.attributes();
	for(unsigned int j=0; j<atts.length(); ++j) {
	  QDomAttr a = atts.item(j).toAttr();
	  if(!a.isNull()) {
	    if(a.name() != "name" && a.name() != "query") {
	      // Bind that sucker
	      q.bindValue(":"+a.name(), a.value());
	      //	      cerr << "Binding " << a.name() << " as " << a.value() << "\n";
	    }
	  }
	}
	q.setForwardOnly(true);
	q.exec();
	if(!q.isActive()) {
	  cerr << "Query failed: " << q.lastError().text() << "\n";
	  exit(1);
	}
	if(q.isSelect()) {
	  QSqlRecordInfo info = db->recordInfo(q);
	  
	  QDomNodeList kids=element.childNodes();
	  bool firstkid = true;
	  int row=0;
	  while(q.next()) {
	    QMap<QString,QString> variable;
	    int col=0;
	    for(QSqlRecordInfo::const_iterator cit=info.begin(); cit!=info.end(); ++cit, ++col) {
	      variable[prefix + (*cit).name() + "$"] = q.value(col).toString();
	    }
	    variable[prefix + "row$"] = QString::number(row++);
	    for(int j=0; j<(int)kids.count(); j++) {
	      QDomNode kid = kids.item(j).cloneNode();
	      // Do variable replace
	      //	    cerr << "Replacing variables\n";
	      replacevariables(kid, variable);
	      if(firstkid) {
		nextnode = kid;
		firstkid=false;
	      }
	      parent.insertBefore(kid, element);
	    }
	  }
	}
	parent.removeChild(element);
      } else if(element.tagName() == "loop") {
	QString sstart = element.attribute("start", "1");
	QString scount = element.attribute("count");
	QString send = element.attribute("end");
	QString sstep = element.attribute("step", "1");
	QString name = element.attribute("name");
	if((scount.isEmpty() && send.isEmpty()) || (!scount.isEmpty() && !send.isEmpty())) {
	  cerr << "<loop> requires either count or end attributes\n";
	  exit(1);
	}
	bool ok;
	int start = sstart.toInt(&ok);
	if(!ok) {
	  cerr << "Bad start value in <loop>: '" << sstart << "'\n";
	  exit(1);
	}
	int step = sstep.toInt(&ok);
	if(!ok) {
	  cerr << "Bad step value in <loop>: '" << sstep << "'\n";
	  exit(1);
	}
	int end;
	if(!send.isEmpty()) {
	  end = send.toInt(&ok);
	  if(!ok) {
	    cerr << "Bad end value in <loop>: '" << send << "'\n";
	    exit(1);
	  }
	} else {
	  int count = scount.toInt(&ok);
	  if(!ok) {
	    cerr << " Bad count value in <loop>: '" << scount << "'\n";
	    exit(1);
	  }
	  end = start + count * step;
	}

	QMap<QString,QString> variable;
	QDomNodeList kids=element.childNodes();
	bool firstkid = true;
	for(int i=start; i<end; i+=step) {
	  if(!name.isEmpty()) {
	    variable["$" + name + ".row$"] = QString::number(i);
	  }
	  for(int j=0; j<(int)kids.count(); j++) {
	    QDomNode kid = kids.item(j).cloneNode();
	    // Do variable replace
	    //	    cerr << "Replacing variables\n";
	    replacevariables(kid, variable);
	    if(firstkid) {
	      nextnode = kid;
	      firstkid=false;
	    }
	    parent.insertBefore(kid, element);
	  }
	}
	parent.removeChild(element);
      } else {
	expandwalk(element);
      }
    }
    node = nextnode;
  }
}

void expanddoc(const QDomElement &parent)
{
  QDomNode node = parent.firstChild();
  while(!node.isNull()) {
    if(node.isElement()) {
      QDomElement el = node.toElement();
      //      cerr << "Expanding " << el.tagName() << "\n";
      replacevariables(el, globalvar);
      expandwalk(el);
    }
    node = node.nextSibling();
  }
  //replacevariables(parent, globalvar);
  //expandwalk(parent);
}

void processfile(QString fn)
{
  //  cerr << "Starting " << fn << "\n";
  QFile xmlfile(fn);
  if(!xmlfile.open(IO_ReadOnly)) {
    cerr << "Failed to open file " << fn << "\n";
    exit(1);
  }
  //  cerr << "Opened file\n";
  QDomDocument doc;
  if(!doc.setContent(&xmlfile)) {
    cerr << "Failed to parse " << fn << "\n";
    xmlfile.close();
    exit(1);
  }
  //  cerr << "Parsed file\n";
  xmlfile.close();
  
  //  walkdoc(doc.documentElement(), 0);
  //  cerr << "Expanding\n";
  QDomElement rootel = doc.documentElement();
  expandwalk(rootel);
  //  expanddoc(doc.documentElement());
 
  //  walkdoc(doc.documentElement(), 0);

  //  cerr << doc.toString() << "\n";
  //  cerr << "Generating doc\n";
  if(writepdf) {
    pdfdocument d;
    gendoc(doc.documentElement(), d);
    d.render(stdout);
  }
  if(writexml) {
    cout << doc.toString(2) << "\n";
  }
}

void initdata()
{
  halign["left"] = pdf::hLeft;
  halign["right"] = pdf::hRight;
  halign["center"] = pdf::hCentre;
  halign["centre"] = pdf::hCentre;
  halign["justify"] = pdf::hJustify;
}
