// -*-C++-*-
// pdfContent.cc
// $Id$
//
// Copyright (C) 2001 Steve Atkins <steve@blighty.com>
// Last change: Time-stamp: <25-Jul-2004 18:40:01 steve>
//
//

#include "pdf.h"
#include "pdfContent.h"

pdfContent::pdfContent(pdfdocument *doc) : pdfchunk(doc),  rawdata(100), dataused(0)
{
}

pdfContent::~pdfContent()
{
}

QByteArray pdfContent::render()
{
  //  cerr << "Entering pdfContent::render()\n";
  QByteArray rendered(dataused + 100);
  char *p = rendered.data() + sprintf(rendered.data(), "<< /Length %d >>\nstream\nq\n", dataused+4);
  qmemmove(p, rawdata.data(), dataused);
  p += dataused;
  p += sprintf(p, "Q\n\nendstream\n");
  rendered.resize(p-rendered.data());
  return rendered;
}

void pdfContent::Do(QCString xobject)
{
  expand(xobject.length() + 10);
  dataused+=sprintf(rawdata.data()+dataused, "/%s Do\n", xobject.data());
}

void pdfContent::TJ(TJtext *text)
{
  TJtext *t = text;
  int headroom=0;
  while(t->kernright != 0.0) {
    headroom += (10 + 2 * t->s.length());
    ++t;
  }
  headroom += (10 + 2 * t->s.length());

  expand(headroom);
  char *p = rawdata.data() + dataused;
  t = text;
  while(t->kernright != 0.0) {
    p+=sprintf(p, "%s %g ", makestring(t->s).data(), t->kernright);
  }
  p+=sprintf(p, "%s TJ\n", makestring(t->s).data());
  dataused = (p - rawdata.data());
}

void pdfContent::TJ(QValueList<TJtext> tx)
{
  int headroom=0;
  QValueList<TJtext> t;
  TJtext pending;
  pending.s = "";
  for(QValueList<TJtext>::const_iterator cit=tx.begin(); cit!=tx.end(); ++cit) {
    //    cerr << "In TJ: " << (*cit).s << " " << (*cit).kernright <<  "\n";
    headroom += (10 + 2 * (*cit).s.length());
    pending.s += (*cit).s;
    if((*cit).kernright > 0.01 || (*cit).kernright < -0.01) {
      pending.kernright = (*cit).kernright;
      t.append(pending);
      pending.s = "";
    }
  }
  if(pending.s != "") {
    pending.kernright = 0.0;
    t.append(pending);
  }
  expand(headroom);

  char *p = rawdata.data() + dataused;
  p+=sprintf(p, "[ ");
  for(QValueList<TJtext>::const_iterator cit=t.begin(); cit!=t.end(); ++cit) {
    p+=sprintf(p, "%s ", makestring((*cit).s).data());
    if((*cit).kernright != 0.0) {
      p+=sprintf(p, "%g ", (*cit).kernright);
    }
  }
  p+=sprintf(p, "] TJ\n");
  dataused = (p - rawdata.data());
}

QCString pdfContent::makestring(const QCString in) const
{
  QCString ret(in.size() * 2 + 2);
  const char *q = in.data();
  char *p = ret.data();
  *p++ = '(';
  char c;
  while((c = *q++)) {
    if(c == '\\' || c == '(' || c == ')')
      *p++ = '\\';
    *p++ = c;
  }
  *p++ = ')';
  *p = '\0';
  return ret;
}
