// -*-C++-*-
// pdfContent.h
// $Id$
//
// Copyright (C) 2001 Steve Atkins <steve@blighty.com>
// Last change: Time-stamp: <13-Jul-2004 17:05:01 steve>
//
//

#ifndef PDFCONTENT_H
#define PDFCONTENT_H

#include "pdfchunk.h"
#include <qstring.h>
#include <qstringlist.h>
#include <qcolor.h>
class pdfdocument;

struct TJtext {
  double kernright;
  QCString s;
};

#define STATICOP(X) {expand(sizeof(X)); qmemmove(rawdata.data()+dataused, X "\n", sizeof(X)); dataused += (sizeof(X));}
#define OP1(X, Y, A) {expand(X); dataused+=sprintf(rawdata.data()+dataused, Y "\n", A);}
#define OP2(X, Y, A, B) {expand(X); dataused+=sprintf(rawdata.data()+dataused, Y "\n", A, B);}
#define OP3(X, Y, A, B, C) {expand(X); dataused+=sprintf(rawdata.data()+dataused, Y "\n", A, B, C);}
#define OP4(X, Y, A, B, C, D) {expand(X); dataused+=sprintf(rawdata.data()+dataused, Y "\n", A, B, C, D);}
#define OP5(X, Y, A, B, C, D, E) {expand(X); dataused+=sprintf(rawdata.data()+dataused, Y "\n", A, B, C, D, E);}
#define OP6(X, Y, A, B, C, D, E, F) {expand(X); dataused+=sprintf(rawdata.data()+dataused, Y "\n", A, B, C, D, E, F);}

class pdfContent : public pdfchunk
{
public:
  pdfContent(pdfdocument *doc);
  virtual ~pdfContent();

  virtual const char * isA() const { return "pdfContent"; };
  virtual QByteArray render();
  void expand(unsigned int size) {while(dataused + size > rawdata.size()) rawdata.resize(rawdata.size() * 2); };
  QCString makestring(const QCString in) const;

  // Stroke functions
  void m(double x, double y) OP2(50, "%g %g m", x, y);
  void l(double x, double y) OP2(50, "%g %g l", x, y);
  void c(double x1, double y1, double x2, double y2, double x3, double y3) OP6(130, "%g %g %g %g %g %g c", x1, y1, x2, y2, x3, y3);
  void v(double x1, double y1, double x2, double y2) OP4(90, "%g %g %g %g v", x1, y1, x2, y2);
  void y(double x1, double y1, double x2, double y2) OP4(90, "%g %g %g %g y", x1, y1, x2, y2);
  void h() STATICOP("h");
  void re(double x, double y, double w, double h) OP4(90, "%g %g %g %g re", x, y, w, h);
  // paint
  void S() STATICOP("S");
  void s() STATICOP("s");
  void f() STATICOP("f");
  void F() STATICOP("F");
  void f_() STATICOP("f*");
  void B() STATICOP("B");
  void B_() STATICOP("B*");
  void b() STATICOP("b");
  void b_() STATICOP("b*");
  void n() STATICOP("n");
  // clip
  void W() STATICOP("W");
  void W_() STATICOP("W*");
  // Color
  void G(double grey) OP1(30, "%g G", grey);
  void g(double grey) OP1(30, "%g g", grey);
  void RG(double red, double green, double blue) OP3(70, "%g %g %g RG", red, green, blue);
  void rg(double red, double green, double blue) OP3(70, "%g %g %g rg", red, green, blue);
  void RG(QColor c) {RG(c.red() / 255.0, c.green() / 255.0, c.blue() / 255.0);};
  void rg(QColor c) {rg(c.red() / 255.0, c.green() / 255.0, c.blue() / 255.0);};
  // Image
  void Do(QCString xobject);
  void cm(double a, double b, double c, double d, double e, double f) OP6(130, "%g %g %g %g %g %g cm",a, b, c, d, e, f);
  void w(double b) OP1(30, "%g w", b);
  // Text
  void BT() STATICOP("BT");
  void ET() STATICOP("ET");
  void Tf(QCString name, double points) OP2(100, "/%s %g Tf", name.data(), points);
  void Tf(int fontid, double points) OP2(100, "/F%d %g Tf", fontid, points);
  void Tc(double space) OP1(30, "%g Tc", space);
  void Tw(double space) OP1(30, "%g Tw", space);
  void Tz(double scale) OP1(30, "%g Tz", scale);
  void TL(double lead) OP1(30, "%g Tf", lead);
  void Tr(int render) OP1(30, "%d Tr", render);
  void Ts(double rise) OP1(30, "%g Ts", rise);
  void Td(double x, double y) OP2(50, "%g %g Td", x, y); // nl
  void TD(double x, double y) OP2(50, "%g %g TD", x, y); // nl + lead
  void Tm(double a, double b, double c, double d, double e, double f) OP6(130, "%g %g %g %g %g %g Tm",a, b, c, d, e, f);
  void T_() STATICOP("T*");
  void Tj(QCString s) OP1(10+s.length(), "%s Tj", makestring(s).data());
  void TJ(TJtext *t);
  void TJ(QValueList<TJtext>);
  void q() STATICOP("q");
  void Q() STATICOP("Q");

  void image(QCString uid, double x, double y, double w, double h) { q(); cm(1, 0, 0, 1, x, y); cm(w, 0, 0, h, 0, 0); Do(uid); Q();};
  void raw(const char *s) {int slen=strlen(s); expand(slen+1); qmemmove(rawdata.data()+dataused, s, slen+1); dataused+=slen;};

private:
  QByteArray rawdata;
  unsigned int dataused;
};


#endif // PDFCONTENT_H
