// -*-C++-*-
// pdfFont.cc
// $Id$
//
// Copyright (C) 2001 Steve Atkins <steve@blighty.com>
// Last change: Time-stamp: <11-Jul-2004 21:07:22 steve>
//
//
#include "pdf.h"
#include "pdfFont.h"

struct charwidthData { 
  int code; 
  int width;
};

struct kernData {
  int left;
  int right;
  int kern;
};

struct fontData {
  const char *fontname;
  const char *fullname;
  const char *familyname;
  int underlineposition;
  int underlinethickness;
  int capheight;
  int xheight;
  int ascender;
  int descender;
  int boxl;
  int boxr;
  int boxt;
  int boxb;
  const charwidthData *charwidth;
  const kernData *kern;
};

#include "afm.h"

// Standard fonts only. But we can extend this easily to support other
// fonts, I think.

pdfFont::pdfFont(pdfdocument *doc, QCString basefont) : pdfchunk(doc), p_basefont(basefont)
{
  const fontData *thisfont = staticfont;
  while(thisfont->fontname && basefont != thisfont->fontname) {
    //    cerr << "Lookin' at " << thisfont->fontname << "\n";
    thisfont++;
  }
  if(!thisfont->fontname) {
    // Not one of the standard fonts, so bail
    THROW(basefont + " Not a standard font" );
  }
  p_underlineposition = thisfont->underlineposition;
  p_underlinethickness = thisfont->underlinethickness;
  p_capheight = thisfont->capheight;
  p_xheight = thisfont->xheight;
  p_ascender = thisfont->ascender;
  p_descender = thisfont->descender;
  p_boxl = thisfont->boxl;
  p_boxr = thisfont->boxr;
  p_boxt = thisfont->boxt;
  p_boxb = thisfont->boxb;
  for(int i=0; i<256; ++i)
    p_width[i] = 0;
  const charwidthData *w = thisfont->charwidth;
  while(w->code) {
    p_width[w->code] = w->width;
    ++w;
  }
  const kernData *k = thisfont->kern;
  while(k->left) {
    p_kern[(k->left << 8) | k->right] = k->kern;
    ++k;
  }
}

pdfFont::~pdfFont()
{
}

QByteArray pdfFont::render()
{
  //  cerr << "Entering pdfFont::render()\n";
  QByteArray ret(1000);
  ret.resize(sprintf(ret.data(), "<< /Type /Font\n/Subtype /Type1\n/BaseFont /%s\n>>\n", p_basefont.data()));
  return ret;
}

double pdfFont::kern(double pointsize, int left, int right) const
{
  if(left<0 || left>255 || right<0 || right>255)
    return 0.0;
  QMap<int,int>::const_iterator cit=p_kern.find((left << 8) | right);
  return cit==p_kern.end() ? 0.0 : cit.data() * pointsize / 1000.0;
}
