// -*-C++-*-
// pdfGrid.cc
// $Id$
//
// Copyright (C) 2001 Steve Atkins <steve@blighty.com>
// Last change: Time-stamp: <25-Jul-2004 19:20:05 steve>
//
//

#include "pdfGrid.h"
#include "pdfPage.h"

pdfGrid::pdfGrid() : nrows(0), ncols(0)
{
}

pdfGrid::~pdfGrid()
{
  // Free all cells
}

pdfCell::pdfCell() : p_width(-1.0), p_height(-1.0)
{
  setPad(4.0);
  setBorder(0.0);
  setCellColor(QColor("white"));
  setWidth(padLeft() + padRight() + borderLeft() + borderRight() + content.width());
  setHeight(padTop() + padBottom() + borderTop() + borderBottom() + content.height());
}

pdfCell::pdfCell(pdfText t) : content(t), p_width(-1.0), p_height(-1.0)
{
  setPad(4.0);
  setBorder(0.0);
  setCellColor(QColor("white"));
  setWidth(padLeft() + padRight() + borderLeft() + borderRight() + content.width());
  setHeight(padTop() + padBottom() + borderTop() + borderBottom() + content.height());
}

void pdfCell::wrapContent(double w)
{
  //  cerr << "pdfCell::wrapContent\n";
  setWidth(w);
  content.wrap(w - padLeft() - padRight() - borderLeft() - borderRight());
  setHeight(padTop() + padBottom() + borderTop() + borderBottom() + content.height());
}

void pdfCell::calculateSize()
{
  setWidth(padLeft() + padRight() + borderLeft() + borderRight() + content.width());
  setHeight(padTop() + padBottom() + borderTop() + borderBottom() + content.height());
}

pdfCellData::pdfCellData(pdfCell *c, int row, int col, int rows, int cols) : p_cell(c), p_row(row), p_col(col), p_rows(rows), p_cols(cols)
{
}

pdfGrid::pdfGrid(double mw) : maxwidth(mw)
{
}

void pdfGrid::wrap(double maxwidth)
{
  DEBUGGRID() { cerr << "pdfGrid::wrap - single cells\n"; }
  QMap<int,bool> cellwidth;
  for(QMap<int,QMap<int,pdfCellData> >::const_iterator yit=grid.begin(); yit!=grid.end(); ++yit) {
    for(QMap<int,pdfCellData>::const_iterator xit=yit.data().begin(); xit!=yit.data().end(); ++xit) {
      if(xit.data().primary()) {
	if(1==xit.data().cols()) {
	  
	  xit.data().cell()->calculateSize();
	  double w = xit.data().cell()->width();
	  if(w > desiredwidth[xit.key()]) {
	    desiredwidth[xit.key()] = w;
	  }
	  DEBUGGRID() {cerr << "    " << xit.key() << "," << yit.key() << " width=" << w << " desiredwidth[" << xit.key() << "] = " << desiredwidth[xit.key()] << "\n";}
	} else {
	  cellwidth[xit.data().cols()] = true;
	}
      }
    }
  }

  for(int cellspan=2; cellspan <= ncols; ++cellspan) {
    DEBUGGRID() { cerr << "pdfGrid::wrap - " << cellspan << " wide cells\n"; }
    if(cellwidth.contains(cellspan)) {
      for(QMap<int,QMap<int,pdfCellData> >::const_iterator yit=grid.begin(); yit!=grid.end(); ++yit) {
	for(QMap<int,pdfCellData>::const_iterator xit=yit.data().begin(); xit!=yit.data().end(); ++xit) {

	  if(xit.data().primary() && xit.data().cols() == cellspan) {
	    int nundef=0;
	    double w=xit.data().cell()->width();
	    double existingwidth=0.0;
	    for(int col=xit.key(); col<xit.key() + cellspan; ++col) {
	      QMap<int,double>::const_iterator fit=desiredwidth.find(col);
	      if(fit == desiredwidth.end()) {
		++nundef;
	      } else {
		existingwidth += fit.data();
	      }
	    }
	    if(w > existingwidth) {
	      if(nundef > 0) {
		double delta = (w-existingwidth) / nundef;
		for(int col=xit.key(); col<xit.key() + cellspan; ++col) {
		  if(!desiredwidth.contains(col)) {
		    desiredwidth[col] = delta;
		  }
		}
	      } else {
		double delta = (w-existingwidth) / cellspan;
		for(int col=xit.key(); col<xit.key() + cellspan; ++col) {
		  desiredwidth[col] += delta;
		}
	      }
	    }
	    DEBUGGRID() {cerr << "    " << xit.key() << "," << yit.key() << " width=" << w << " desiredwidth[" << xit.key() << "] = " << desiredwidth[xit.key()] << "\n";}

	  }
	}
      }
    }
  }

  double colmax = maxwidth / ncols;
  int nwidecols=0;
  double dw=0.0;
  double fw=0.0;
  for(QMap<int,double>::const_iterator cit=desiredwidth.begin(); cit!=desiredwidth.end(); ++cit) {
    DEBUGGRID() {cerr << "Desired: " << cit.key() << " -> " << cit.data() << "\n";}
    dw += cit.data();
    if(cit.data() > colmax) {
      nwidecols++;
    } else {
      fw += cit.data();
    }
  }
  if(dw <= maxwidth) {
    columnwidth = desiredwidth;
  } else {
    while(nwidecols && colmax != (maxwidth-fw)/nwidecols) {
      DEBUGGRID() {cerr << "Entering column expansion loop\n";}
      colmax = (maxwidth - fw) / nwidecols;
      fw = 0.0;
      nwidecols = 0;
      for(QMap<int,double>::const_iterator cit=desiredwidth.begin(); cit!=desiredwidth.end(); ++cit) {
	if(cit.data() > colmax) {
	  columnwidth[cit.key()] = colmax;
	  nwidecols++;
	} else {
	  columnwidth[cit.key()] = cit.data();
	  fw += cit.data();
	}
	DEBUGGRID() {cerr << "Final: " << cit.key() << " -> " << columnwidth[cit.key()] << "\n";}
      }
    }

    //    cerr << "About to wrap cells\n";
    DEBUGGRID() {cerr << "Wrapping cells\n";}
    QMap<int,bool> rowspan;
    for(QMap<int,QMap<int,pdfCellData> >::const_iterator yit=grid.begin(); yit!=grid.end(); ++yit) {
      for(QMap<int,pdfCellData>::const_iterator xit=yit.data().begin(); xit!=yit.data().end(); ++xit) {
	if(xit.data().primary()) {
	  double w=0.0;
	  for(int col=xit.key(); col < xit.key() + xit.data().cols(); ++col ) {
	    w += columnwidth[col];
	  }
	  xit.data().cell()->wrapContent(w);
	  if(xit.data().rows() == 1) {
	    double nh =  xit.data().cell()->height();
	    if(nh > rowheight[yit.key()])
	      rowheight[yit.key()] = nh;
	  } else {
	    rowspan[xit.data().rows()] = true;
	  }
	}
      }
    }
    
    for(int rs=2; rs<nrows; ++rs) {
      if(rowspan.contains(rs)) {
	for(QMap<int,QMap<int,pdfCellData> >::const_iterator yit=grid.begin(); yit!=grid.end(); ++yit) {
	  for(QMap<int,pdfCellData>::const_iterator xit=yit.data().begin(); xit!=yit.data().end(); ++xit) {
	    if(xit.data().primary() && xit.data().rows() == rs) {
	      double currheight = 0;
	      for(int row=yit.key(); row < yit.key() + xit.data().rows(); ++row) {
		currheight += rowheight[row];
	      }
	      if(xit.data().cell()->height() > currheight) {
		double delta = (xit.data().cell()->height() - currheight) / xit.data().rows();
		for(int row=yit.key(); row < yit.key() + xit.data().rows(); ++row) {
		  rowheight[row] += delta;
		}
	      }
	    }
	  }
	}
      }
    }
  }

  for(QMap<int,QMap<int,pdfCellData> >::const_iterator yit=grid.begin(); yit!=grid.end(); ++yit) {
    for(QMap<int,pdfCellData>::const_iterator xit=yit.data().begin(); xit!=yit.data().end(); ++xit) {
      xit.data().cell()->wrapContent(columnwidth[xit.key()]);
      //      DEBUGGRID() {cerr << "Wrapping " << xit.key() << "," << yit.key() << " -> " << columnwidth[xit.key()] << "\n";}
    }
  }

  recurrentheight = 0.0;
  for(QMap<int,QMap<int,pdfCellData> >::const_iterator yit=grid.begin(); yit!=grid.end(); ++yit) {
    double maxheight=0.0;
    
    for(QMap<int,pdfCellData>::const_iterator xit = yit.data().begin(); xit!=yit.data().end(); ++xit) {
      if(xit.data().primary() && xit.data().cell()->height() > maxheight)
	maxheight = xit.data().cell()->height();
    }
    rowheight[yit.key()] = maxheight;
    if(persistentRow(yit.key()))
      recurrentheight += maxheight;
  }

  // Force all cells in a row to be same height
  for(QMap<int,QMap<int,pdfCellData> >::const_iterator yit=grid.begin(); yit!=grid.end(); ++yit) {
    double maxheight=0.0;
    for(QMap<int,pdfCellData>::const_iterator xit = yit.data().begin(); xit!=yit.data().end(); ++xit) {
      if(xit.data().primary() && xit.data().rows() == 1 && xit.data().cell()->height() > maxheight) {
	maxheight = xit.data().cell()->height();
      }
    }
    for(QMap<int,pdfCellData>::const_iterator xit = yit.data().begin(); xit!=yit.data().end(); ++xit) {
      if(xit.data().primary() && xit.data().rows() == 1) {
	xit.data().cell()->setHeight(maxheight);
      }
    }
  }
}

int pdfGrid::slice(double maxheight, int startrow, bool persistent)
{
  if(persistent)
    maxheight -= recurrentheight;

  while(startrow < nrows) {
    QMap<int,double>::const_iterator cit = rowheight.find(startrow);
    if(cit != rowheight.end() && (!persistent || !persistentRow(startrow))) {
      if(cit.data() > maxheight)
	return startrow;
      maxheight -= cit.data();
    }
    ++startrow;
  }
  return startrow;
}

double pdfGrid::writeSlice(pdfPage *p, int startrow, int endrow, double x, double y, bool persistent) const
{
  double height=0.0;
  pdfContent *c = p->newContent();
  for(int row=0; row < nrows; ++row) {
    if((persistent && persistentRow(row)) || (row >= startrow && row < endrow)) {
      writeRow(p, c, row, x, y-height);
      height += rowHeight(row);
    }
  }
  return height;
}

void pdfGrid::insertCell(const pdfCell& c, int x, int y, int w, int h)
{
  //  cerr << "+++ insertCell(" << x << "," << y << ")\n";
  if(x+w > ncols)
    ncols = x+w;
  if(y+h > nrows)
    nrows = y+h;

  pdfCell *cell = new pdfCell(c);

  for(int nx=0; nx<w; nx++) {
    for(int ny=0; ny<h; ny++) {
      pdfCellData cd(cell, ny, nx, h, w);
      grid[y+ny][x+nx] = cd;
    }
  }
}

void pdfGrid::writeRow(pdfPage *p, pdfContent *c, int row, double x, double y) const
{
  //  cerr << "+ writeRow " << row << " " << x << " " << y << "\n";
  QMap<int,QMap<int,pdfCellData> >::const_iterator yit=grid.find(row);
  if(yit == grid.end())
    return;
  
  if(0==c)
    c = p->newContent();

  double nx=x;

  for(QMap<int,pdfCellData>::const_iterator xit = yit.data().begin(); xit!=yit.data().end(); ++xit) {
    if(xit.data().primary()) {
      xit.data().cell()->writebg(c, nx, y);
      nx += xit.data().cell()->width();
    }
  }
  nx = x;

  for(QMap<int,pdfCellData>::const_iterator xit = yit.data().begin(); xit!=yit.data().end(); ++xit) {
    if(xit.data().primary()) {
      xit.data().cell()->writefg(p, nx, y);
      nx += xit.data().cell()->width();
    }
  }
}

void pdfGrid::dump()
{
  for(QMap<int,QMap<int,pdfCellData> >::const_iterator yit = grid.begin(); yit != grid.end(); ++yit) {
    for(QMap<int,pdfCellData>::const_iterator xit = yit.data().begin(); xit!=yit.data().end(); ++xit) {
      //      cerr << "grid " << xit.key() << "," << yit.key() << " width=" << xit.data().cell()->width() << " height=" << xit.data().cell()->height() << "\n";
    }
  }
}

void pdfCell::writefg(pdfPage *p, double x, double y) const
{
  //  cerr << "pdfCell::writefg x=" << x << " y=" << y << " pl=" << padLeft() << " bl=" << borderLeft() << "\n";
  //  content.dumpsize();
  //  cerr << "my width=" << p_width << " height=" << p_height << "\n";
  content.write(p, x+padLeft()+borderLeft(), y-padTop()-borderTop());
}

void pdfCell::writebg(pdfContent *c, double x, double y) const
{
  // if(borderTop() && borderTop() == borderLeft() && borderTop() == borderRight() && borderTop() == borderBottom() && borderTopColor() == borderLeftColor() && borderTopColor() == borderRightColor() && ) {
  //  }

  c->rg(cellColor()); // RGB mode
  c->re(x, y, width(), -height()); // Rectangle
  c->f();

  if(borderTop()) {
    c->RG(borderTopColor());
    c->w(borderTop());
    c->m(x, y);
    c->l(x+width(), y);
    c->s();
  }
  if(borderLeft()) {
    c->RG(borderLeftColor());
    c->w(borderLeft());
    c->m(x, y);
    c->l(x, y-height());
    c->s();
  }
  if(borderRight()) {
    c->RG(borderRightColor());
    c->w(borderRight());
    c->m(x+width(), y);
    c->l(x+width(), y-height());
    c->s();
  }
  if(borderBottom()) {
    c->RG(borderBottomColor());
    c->w(borderBottom());
    c->m(x, y-height());
    c->l(x+width(), y-height());
    c->s();
  }
}
