// -*-C++-*-
// pdfGrid.h
// $Id$
//
// Copyright (C) 2001 Steve Atkins <steve@blighty.com>
// Last change: Time-stamp: <05-Jul-2004 19:43:57 steve>
//
//

#ifndef PDFGRID_H
#define PDFGRID_H

#include <qcolor.h>
#include <qmap.h>
#include "pdfText.h"
#include "pdfContent.h"

class pdfCell
{
public:
  pdfCell();
  pdfCell(pdfText t);
  virtual ~pdfCell() {};
  virtual const char * isA() const { return "pdfCell"; };

  double desiredWidth() {return padLeft() + padRight() + borderLeft() + borderRight() + content.width(); };

  void wrapContent(double w);

  void setWidth(double b) {p_width=b;};
  void setHeight(double b) {p_height=b;};
  
  double height() const {return p_height;};
  double width() const {return p_width;};

  void setPadTop(double b) {p_pad_top=b;};
  void setPadLeft(double b) {p_pad_left=b;};
  void setPadRight(double b) {p_pad_right=b;};
  void setPadBottom(double b) {p_pad_bottom=b;};
  void setPad(double b) {setPadTop(b);setPadLeft(b);setPadRight(b);setPadBottom(b);};
  void setBorderTop(double b) {p_border_top=b;};
  void setBorderLeft(double b) {p_border_left=b;};
  void setBorderRight(double b) {p_border_right=b;};
  void setBorderBottom(double b) {p_border_bottom=b;};
  void setBorder(double b) {setBorderTop(b);setBorderLeft(b);setBorderRight(b);setBorderBottom(b);};

  double padTop() const {return p_pad_top;};
  double padLeft() const {return p_pad_left;};
  double padRight() const {return p_pad_right;};
  double padBottom() const {return p_pad_bottom;};

  double borderTop() const {return p_border_top;};
  double borderLeft() const {return p_border_left;};
  double borderRight() const {return p_border_right;};
  double borderBottom() const {return p_border_bottom;};

  void setCellColor(QColor b) {p_col_cell=b;};
  void setCellColor(const QString& b) {p_col_cell.setNamedColor(b);};
  void setBorderTopColor(QColor b) {p_col_border_top=b;};
  void setBorderTopColor(const QString& b) {p_col_border_top.setNamedColor(b);};
  void setBorderLeftColor(QColor b) {p_col_border_left=b;};
  void setBorderLeftColor(const QString& b) {p_col_border_left.setNamedColor(b);};
  void setBorderRightColor(QColor b) {p_col_border_right=b;};
  void setBorderRightColor(const QString& b) {p_col_border_right.setNamedColor(b);};
  void setBorderBottomColor(QColor b) {p_col_border_bottom=b;};
  void setBorderBottomColor(const QString& b) {p_col_border_bottom.setNamedColor(b);};
  void setBorderColor(QColor b) {setBorderTopColor(b);setBorderLeftColor(b);setBorderRightColor(b);setBorderBottomColor(b); }
  void setBorderColor(const QString& b) {setBorderTopColor(b);setBorderLeftColor(b);setBorderRightColor(b);setBorderBottomColor(b); }

  QColor cellColor() const {return p_col_cell;};
  QColor borderTopColor() const {return p_col_border_top;};
  QColor borderLeftColor() const {return p_col_border_left;};
  QColor borderRightColor() const {return p_col_border_right;};
  QColor borderBottomColor() const {return p_col_border_bottom;};

  pdfText content;

  void writefg(pdfPage *p, double x, double y) const;
  void writebg(pdfContent *c, double x, double y) const;
  void calculateSize();
private:
  double p_width;
  double p_height;
  
  double p_pad_top;
  double p_pad_left;
  double p_pad_right;
  double p_pad_bottom;
  double p_border_top;
  double p_border_left;
  double p_border_right;
  double p_border_bottom;
  
  QColor p_col_cell;
  QColor p_col_border_top;
  QColor p_col_border_left;
  QColor p_col_border_right;
  QColor p_col_border_bottom;

};

class pdfCellData {
public:
  pdfCellData() {};
  pdfCellData(pdfCell *c, int row, int col, int rows, int cols);
  bool primary() const {return 0==p_row && 0==p_col;};
  int rows() const {return p_rows;};
  int cols() const {return p_cols;};
  pdfCell *cell() const {return p_cell;};
private:
  pdfCell *p_cell;
  int p_row;
  int p_col;
  int p_rows;
  int p_cols;
};

class pdfGrid
{
public:
  pdfGrid();
  pdfGrid(double mw);
  virtual ~pdfGrid();

  virtual const char * isA() const { return "pdfGrid"; };
  void wrap(double maxwidth);
  void insertCell(const pdfCell& c, int x, int y, int w, int h);
  void setPersistentRow(int row, bool b=true) {recurrent[row]=b;};
  bool persistentRow(int row) const {QMap<int,bool>::const_iterator c=recurrent.find(row); return c==recurrent.end() ? false : c.data();};


  int numRows() const { return nrows; };
  double rowHeight(int row) const { return rowheight[row]; };
  void writeRow(pdfPage *p, pdfContent *c, int row, double x, double y) const;
  void dump();
  int rows() const { return nrows; };
  int slice(double maxheight, int startrow, bool persistent=true);
  double writeSlice(pdfPage *p, int startrow, int endrow, double x, double y, bool persistent=true) const;
private:
  QMap<int,double> rowheight;
  QMap<int,double> columnwidth;
  QMap<int,double> desiredwidth;
  QMap<int,bool> recurrent;
  double recurrentheight;

  double maxwidth;

  int nrows;
  int ncols;
  QMap<int,QMap<int,pdfCellData> > grid; // grid[y][x]
};


#endif // PDFGRID_H
