// -*-C++-*-
// pdfPage.cc
// $Id: pdfPage.cc,v 1.1 2004/06/13 00:26:57 steve Exp steve $
//
// Copyright (C) 2001 Steve Atkins <steve@blighty.com>
// Last change: Time-stamp: <13-Jul-2004 18:09:18 steve>
//
//

#include "pdfPage.h"
#include "pdfdocument.h"
#include "pdfContent.h"
#include "pdfPages.h"
#include "pdfXobject.h"
#include "pdfImage.h"

pdfPage::pdfPage(pdfdocument *doc, double width, double height) : pdfchunk(doc), p_parent(0), p_thumbnail(0), p_height(height), p_width(width), fontspin(0)
{
  pdfContent *c = newContent();
  fontTag("Times-Roman");
}

pdfPage::~pdfPage()
{
}

QCString pdfPage::fontTag(QCString basefont)
{
  QMap<QCString,QCString>::const_iterator cit=fonttag.find(basefont);
  if(cit != fonttag.end())
    return cit.data();
  QCString ret(20);
  sprintf(ret.data(), "F%d", ++fontspin);
  useFont(ret, basefont);
  return ret;
}

pdfContent *pdfPage::newContent()
{
  pdfContent *c = new pdfContent(document());
  includeContent(c);
  return c;
}

void pdfPage::insertImage(const QCString name, const QImage& image)
{
  includeXobject(name, new pdfImage(document(), image));
}

QByteArray pdfPage::render()
{
  //  cerr << "Entering pdfPage::render()\n";
  QByteArray ret(200 + 40 * (fontmap.count() + content.count() + xobject.count()));
  char *p=ret.data();
  p+=sprintf(p, "<< /Type /Page\n/Parent %d 0 R\n", parent()->index());
  p+=sprintf(p, "/Resources <<\n");
  if(!fontmap.isEmpty()) {
    p+=sprintf(p, "/Font << ");
    for(QMap<QCString,QCString>::const_iterator cit=fontmap.begin(); cit!=fontmap.end(); ++cit) {
      p+=sprintf(p, "/%s %d 0 R\n", cit.key().data(), document()->getFontRef(cit.data()));
    }
    p+=sprintf(p, ">>\n");
  }
  if(!xobject.isEmpty()) {
    p+=sprintf(p, "/XObject <<\n");
    for(QMap<QCString,pdfXobject*>::const_iterator cit=xobject.begin(); cit!=xobject.end(); ++cit) {
      p+=sprintf(p, "/%s %d 0 R\n", cit.key().data(), cit.data()->index());
    }
    p+=sprintf(p, ">>\n");
  }
  p+=sprintf(p, "/ProcSet [/PDF /Text /ImageB /ImageC /ImageI]\n>>\n");
  p+=sprintf(p, "/MediaBox [0 0 %f %f]\n", p_width, p_height);
  p+=sprintf(p, "/Contents [\n");
  for(QValueList<pdfContent*>::const_iterator cit = content.begin(); cit!=content.end(); ++cit) {
    p+=sprintf(p, "%d 0 R\n", (*cit)->index());
  }
  p+=sprintf(p, "]\n");
  if(p_thumbnail) {
    p+=sprintf(p, "/Thumb %d 0 R\n", p_thumbnail->index());
  }
  p+=sprintf(p, ">>\n");
  ret.resize(p-ret.data());
  return ret;
}


