// -*-C++-*-
// pdfText.cc
// $Id: pdfText.cc,v 1.1 2004/06/16 00:26:59 steve Exp steve $
//
// Copyright (C) 2001 Steve Atkins <steve@blighty.com>
// Last change: Time-stamp: <11-Jul-2004 21:06:28 steve>
//
//

#include "pdf.h"
#include "pdfText.h"
#include "pdfdocument.h"
#include "pdfPage.h"
#include "pdfContent.h"
#include "pdfFont.h"

pdfTextChunk::pdfTextChunk(const pdfTextChunk& s) : type(s.type), valign(s.valign), width(s.width), heightup(s.heightup), heightdown(s.heightdown), kern(s.kern), /*font(s.font),*/ pointsize(s.pointsize), underline(s.underline), wordpad(s.wordpad), charpad(s.charpad)/*, text(s.text), imageid(s.imageid)*/
{
  text = s.text;
  font = s.font;
  imageid = s.imageid;
}

pdfTextChunk::pdfTextChunk(pdfText *t, pdf::Type tp) : type(tp), valign(t->valign()), width(0.0), heightup(0.0), heightdown(0.0), kern(0.0), pointsize(t->pointsize()), underline(t->underline()), wordpad(t->wordpad()), charpad(t->charpad())
{
  if(t->font() == "Times") {
    if(t->bold()) {
      if(t->italic()) {
	font = t->font() + "-BoldItalic";
      } else {
	font = t->font() + "-Bold";
      }
    } else {
      if(t->italic()) {
	font = t->font() + "-Italic";
      } else {
	font = "Times-Roman";
      }
    }
  } else {
    if(t->bold()) {
      if(t->italic()) {
	font = t->font() + "-BoldOblique";
      } else {
	font = t->font() + "-Bold";
      }
    } else {
      if(t->italic()) {
	font = t->font() + "-Oblique";
      } else {
	font = t->font();
      }
    }
  }
}

pdfText::pdfText(pdfdocument *doc) : p_halign(pdf::hLeft), p_font("Times"), p_italic(false), p_bold(false), p_underline(false), p_pointsize(12.0), p_wordpad(0.0), p_charpad(0.0), p_kern(true), p_valign(pdf::vBaseline), p_document(doc), sizecalculated(false)
{
}

pdfText::~pdfText()
{
}

void pdfTextChunk::calcWidth(const pdfFont *pf)
{
  //  cerr << "Entering pdfTextChunk::calcWidth for '" << text << "'\n";
  const char *p = text.data();
  width = 0.0;
  while(*p) {
    width += pf->width(pointsize, *p);
    if(*p == ' ') {
      width += wordpad;
    }
    width += charpad;
    ++p;
  }
  //  cerr << "width=" << width << "\n";
}

// Enter a single, non-kerned chunk
void pdfText::textchunk(pdfTextChunk c, const pdfFont *pf)
{
  //  cerr << "Entering pdfText::textchunk '" << c.text << "'\n";
  if(c.text.isEmpty())
    return;

  char *p = c.text.data();

  if(p_kern && !chunk.empty()) {
    pdfTextChunk& t=chunk.last();
    if(t.type == pdf::Text && !t.text.isEmpty()) {
      if(t.font == c.font && t.pointsize == c.pointsize) {
	t.kern = pf->kern(c.pointsize, t.text[0], *p);
      } else {
	// Do cross-font kerning XXXX
      }
    }
  }

  double w=0;
  while(*p) {
    //...for each glyph, add width
    w += pf->width(c.pointsize, *p);
    if(*p == ' ') {
      w += c.wordpad;
    }
    w += c.charpad;
    p++;
  }
  c.heightup = pf->ascender(c.pointsize);
  c.heightdown = pf->descender(c.pointsize);
  c.width = w;
  c.kern = 0;
  chunk.append(c);
  //  cerr << "Chunk width=" << w << "\n";
  //  {
  //    for(QValueList<pdfTextChunk>::const_iterator cit=chunk.begin(); cit!=chunk.end(); ++cit) {
  //      cerr << c.text << " Chunk dump " << (*cit).text << "\n";
  //    }
  //  }
}

void pdfText::append(QCString s)
{
  pdfTextChunk c(this, pdf::Text);
  const pdfFont *pf = document()->useFont(c.font);

  //  cerr << "pdfText::append(" << s << ")\n";

  if(!p_kern) {
    textchunk(c, pf);
    c.text = s.copy();
  } else {
    QCString el(s.length() + 2);
    // split string into kernable elements
    const char *q = s.data();
    char *p = el.data();
    while(*q) {
      char ch = *q++;
      *p++ = ch;
      if(*q == '\0')
	break;
      if(pf->isKerned(ch, *q)) {
	*p = '\0';
	c.text = el.copy();
	textchunk(c, pf);
	p = el.data();
      }
    }
    *p = '\0';
    c.text = el.copy();
    textchunk(c, pf);
  }
}

void pdfText::append(QCString uid, QImage im, double width, double height)
{
  //  cerr << "Appending image '" << uid << "'\n";
  document()->append(uid, im);

  pdfTextChunk c(this, pdf::Image);
  c.width = width;
  c.heightup = height;
  c.imageid = uid.copy();
  chunk.append(c);
}

void pdfText::append(QCString uid, double width, double height)
{
  if(!document()->existsXobject(uid)) {
    THROW("Image '" + uid + "' doesn't exist");
  }
  pdfTextChunk c(this, pdf::Image);
  c.width = width;
  c.heightup = height;
  c.imageid = uid.copy();
  chunk.append(c);
}

struct imageplace {
  QCString id;
  double x;
  double y;
  double w;
  double h;
};

void pdfText::write(pdfPage *page, double x, double y) const
{
  QValueList<imageplace> images;
  //  cerr << "pdfText::write x=" << x << " y=" << y << "\n";

  //  {
  //    for(QValueList<pdfTextChunk>::const_iterator cit=chunk.begin(); cit!=chunk.end(); ++cit) {
      //      cerr << "Chunkdump " << (*cit).text << "\n";
  //    }
  //  }
  
  pdfContent *c = new pdfContent(document());

  //  cerr << "write(" << x << "," << y << ")\n";

  double lastx=-x, lasty=-y;
  
  double nx=0.0;
  double ny=0.0;

  QValueList<pdfTextChunk>::const_iterator cit=chunk.begin();  

  QCString font;
  double pointsize(-1.0);

  c->BT();

  while(cit!=chunk.end()) {
    while(cit!=chunk.end() && (*cit).type == pdf::Break) {
      ny -= ((*cit).heightup - (*cit).heightdown);
      ++cit;
    }
    QValueList<pdfTextChunk>::const_iterator t=cit;
    double hup(0.0), hdown(0.0), w(0.0);
    while(t!=chunk.end() && (*t).type != pdf::Break) {
      w += (*t).width;
      if((*t).type == pdf::Text && (*t).kern != 0.0) {
	QValueList<pdfTextChunk>::const_iterator next = t;
	++next;
	if(next != chunk.end() && (*next).type == pdf::Text) {
	  w -= (*t).kern;
	}
      }
      if((*t).heightup > hup)
	hup = (*t).heightup;
      if((*t).heightdown < hdown)
	hdown = (*t).heightdown;
      ++t;
    }
    
    switch(halign()) {
    case pdf::hLeft:
      nx = 0;
      break;
    case pdf::hRight:
      nx = boxwidth() - w;
      break;
    case pdf::hCentre:
      nx = (boxwidth() - w)/2.0;
      break;
    case pdf::hJustify:
      nx = 0;
      break;
    }

    switch(valign()) {
    case pdf::vTop:
      break;
    case pdf::vCentre:
      break;
    case pdf::vBaseline:
      break;
    case pdf::vBottom:
      break;
    }

    ny -= hup;
    t=cit;

    //    cerr << "Moving from " << lastx << "," << lasty << " to " << nx << "," << ny << "\n";
    //    cerr << "hup=" << hup << " hdown=" << hdown << "\n";

    //    cerr << "lastx=" << lastx << " lasty=" << lasty << " nx=" << nx << " ny=" << ny << "\n";
    c->Td(nx-lastx, ny-lasty);
    lastx=nx;
    lasty=ny;
    
    while(t!=chunk.end() && (*t).type != pdf::Break) {
      if((*t).type == pdf::Text) {
	if((*t).font != font || (*t).pointsize != pointsize) {
	  c->Tf(page->fontTag((*t).font), (*t).pointsize);
	  font = (*t).font;
	  pointsize = (*t).pointsize;
	}
	QValueList<TJtext> line;
	//	cerr << "Mark\n";
	while(t!=chunk.end() && (*t).type == pdf::Text && (*t).font == font && (*t).pointsize == pointsize) {
	  TJtext w;
	  w.kernright = (*t).kern * -1000.0 / pointsize;
	  w.s = (*t).text;
	  //	  cerr << "-- " << (*t).text << " " << w.s << " " << w.kernright << "\n";
	  line.append(w);
	  nx += (*t).width - (*t).kern;
	  ++t;
	}
	c->TJ(line);
	continue;
      }
      if((*t).type == pdf::Space) {
	nx += (*t).width;
	c->Td(nx-lastx, ny-lasty);
	lastx=nx;
	lasty=ny;
	++t;
	continue;
      }
      if((*t).type == pdf::Image) {
	imageplace ip;
	ip.id = (*t).imageid;
	ip.x = nx + x;
	ip.y = ny + y;
	ip.w = (*t).width;
	ip.h = (*t).heightup - (*t).heightdown;
	images.append(ip);
	//	page->includeXobject((*t).imageid);
	//	c->image((*t).imageid, nx, ny, (*t).width, (*t).heightup - (*t).heightdown);
	nx += (*t).width;
	c->Td(nx-lastx, ny-lasty);
	lastx=nx;
	lasty=ny;
	++t;
	continue;
      }
    }  

    ny += hdown;

    while(cit!=chunk.end() && (*cit).type != pdf::Break) {
      ++cit;
    }
  }

  c->ET();

  for(QValueList<imageplace>::const_iterator cit=images.begin(); cit!=images.end(); ++cit) {
    page->includeXobject((*cit).id);
    //    cerr << "width = " << (*cit).w << " height = " << (*cit).h << "\n";
    //    cerr << "x = " << (*cit).x << " y = " << (*cit).y << "\n";
    c->image((*cit).id, (*cit).x, (*cit).y, (*cit).w, (*cit).h);
  }

  page->includeContent(c);
}

void pdfText::wrap()
{
  QValueList<pdfTextChunk>::iterator it=chunk.begin();
  double w = 0.0;
  
  while(it!=chunk.end()) {
    //    cerr << "chunktext = '" << (*it).text << "'\n";
    //    cerr << "w=" << w << " chunkwidth=" << (*it).width << "\n";
    if((*it).type == pdf::Break) {
      //      cerr << "Break\n";
      w = 0.0;
      ++it;
      continue;
    }

    ///    cerr << "chunk='" << (*it).text << "' chunkwidth=" << (*it).width << " aggw=" << w << " boxwidth=" << boxwidth() << "\n";

    if(w + (*it).width > boxwidth()+0.001) {
      if((*it).type == pdf::Text) {
	// split off first word.
	int spaceidx = (*it).text.find(' ', 1);
	if(spaceidx == -1 || spaceidx == 0 || spaceidx == (int)((*it).text.length())-1) {
	  // No space
	  if(w==0.0) {
	    // Begining of line
	    if((*it).text.length() < 2) {
	      // Insert single char thing
	      ++it;
	      pdfTextChunk brk(this, pdf::Break);
	      chunk.insert(it, brk);
	      continue;
	    }
	    pdfTextChunk firstchar = *it;
	    firstchar.kern = 0.0;
	    firstchar.text = (*it).text.left(1);
	    (*it).text = (*it).text.mid(1);
	    const pdfFont *pf = document()->useFont(firstchar.font);
	    firstchar.calcWidth(pf);
	    (*it).calcWidth(pf);
	    chunk.insert(it, firstchar);
	    --it;
	  } else {
	    pdfTextChunk brk(this, pdf::Break);
	    chunk.insert(it, brk);
	    w = 0.0;
	    continue;
	  }
	} else {
	  pdfTextChunk firstword = *it;
	  firstword.text = (*it).text.left(spaceidx+1);
	  (*it).text = (*it).text.mid(spaceidx+1);
	  const pdfFont *pf = document()->useFont(firstword.font);
	  firstword.kern = 0.0;
	  firstword.calcWidth(pf);
	  (*it).calcWidth(pf);
	  chunk.insert(it, firstword);
	  --it;
	  continue;
	}
      } else {
	if(w == 0.0) {
	  pdfTextChunk brk(this, pdf::Break);
	  ++it;
	  chunk.insert(it, brk);
	} else {
	  pdfTextChunk brk(this, pdf::Break);
	  chunk.insert(it, brk);
	  w = 0.0;
	}
      }
    } else {
      w += (*it).width;
      ++it;
    }
  }
}

void pdfText::calculatesize()
{
  //  cerr << "Entering pdfText::calculatesize\n";
  p_height = 0.0;
  p_width = 0.0;
  double w=0.0;
  double hup=0.0;
  double hdown=0.0;
  for(QValueList<pdfTextChunk>::const_iterator cit=chunk.begin(); cit!=chunk.end(); ++cit) {
    if((*cit).heightup > hup)
      hup = (*cit).heightup;
    if((*cit).heightdown < hdown)
      hdown = (*cit).heightdown;
    
    w += (*cit).width;
    if((*cit).type == pdf::Break) {
      p_height += (hup - hdown);
      hup = 0.0;
      hdown = 0.0;
      if(w > p_width)
	p_width = w;
      w = 0.0;
    }
  }
  // In case there was no trailing Break
  if(w > p_width)
    p_width = w;
  p_height += (hup - hdown);
  //  cerr << "height=" << p_height << " width=" << p_width << "\n";
}

// Truncates this stream of text to no more than height high, and returns
// remainder
pdfText *pdfText::vsplit(double heightlimit)
{
  //  cerr << "vsplit(" << heightlimit << ")\n";
  //  exit(1);
  p_height = 0.0;
  p_width = 0.0;
  double w=0.0;
  double hup=0.0;
  double hdown=0.0;

  QValueList<pdfTextChunk>::iterator linestart = chunk.begin();
  

  for(QValueList<pdfTextChunk>::iterator cit=chunk.begin(); cit!=chunk.end(); ++cit) {
    QValueList<pdfTextChunk>::iterator nextchunk=cit;
    ++nextchunk;

    if((*cit).heightup > hup)
      hup = (*cit).heightup;
    if((*cit).heightdown < hdown)
      hdown = (*cit).heightdown;
    //    cerr << "hup=" << hup << " hdown=" << hdown << " p_height=" << p_height << "\n";
    if((*cit).type == pdf::Break || nextchunk == chunk.end()) {
      if(w > p_width)
	p_width = w;
      w = 0.0;

      if(p_height + (hup - hdown) > heightlimit) {
	// We're done. Truncate and return and stuff
	pdfText *remainder = new pdfText(document());
	while(linestart != chunk.end()) {
	  remainder->append(*linestart);
	  linestart = chunk.remove(linestart);
	}
	return remainder;
      }

      p_height += (hup - hdown);
      hup = 0.0;
      hdown = 0.0;
      linestart = cit;
      linestart++;
    }
  }
  return 0;
}
