// -*-C++-*-
// pdfText.h
// $Id: pdfText.h,v 1.1 2004/06/16 00:27:01 steve Exp steve $
//
// Copyright (C) 2001 Steve Atkins <steve@blighty.com>
// Last change: Time-stamp: <05-Jul-2004 20:40:10 steve>
//
//

#ifndef PDFTEXT_H
#define PDFTEXT_H

#include <qcstring.h>
#include <qstring.h>
#include <qimage.h>
#include <qcolor.h>
#include "pdf.h"

class pdfFont;
class pdfImage;
class pdfText;
class pdfdocument;
class pdfPage;

using namespace std;

class pdfTextChunk
{
public:
  pdfTextChunk() {};
  pdfTextChunk(const pdfTextChunk& s);
  pdfTextChunk(pdfText *t, pdf::Type tp);
  pdf::Type type;
  pdf::vAlign valign;

  double width;
  double heightup;   // (+ve)
  double heightdown; // (-ve)
  double kern;

  // Text
  QCString font;
  double pointsize;
  bool underline;
  double wordpad;
  double charpad;
  QCString text;

  // Image
  QCString imageid;
  void calcWidth(const pdfFont *pf);
};

class pdfText
{
public:
  pdfText(pdfdocument*);
  pdfText() {};
  virtual ~pdfText();

  virtual const char * isA() const { return "pdfText"; };

  bool setFont(const QCString& font) {p_font = font; return true;};
  void setItalic(bool b) {p_italic = b;};
  void setBold(bool b) {p_bold = b;};
  void setUnderline(bool b) {p_underline = b;};
  void setPointsize(double b) {p_pointsize = b;};
  void setWordpad(double b) {p_wordpad = b;};
  void setCharpad(double b) {p_charpad = b;};
  void setKern(bool b) {p_kern = b;};
  void setVAlign(pdf::vAlign b) {p_valign = b;};
  void setHAlign(pdf::hAlign b) {p_halign = b;};
  void setBoxwidth(double b) {p_boxwidth = b;};

  void clear() { chunk.clear(); };
  void textchunk(pdfTextChunk c, const pdfFont *pf);
  void append(const QCString c);
  void append(QCString uid, QImage c, double width, double height);
  void append(QCString uid, double width, double height);
  void append(pdfTextChunk& c) {chunk.append(c);};
  void appendBreak(double height=0.0) {pdfTextChunk s(this, pdf::Break); s.heightup = height; chunk.append(s);};
  void appendSpace(double width) {pdfTextChunk s(this, pdf::Space); s.width=width; chunk.append(s);};
  pdfdocument *document() const { return p_document;};

  QCString font() const { return p_font; };
  bool italic() const { return p_italic; };
  bool bold() const { return p_bold; };
  bool underline() const { return p_underline; };
  double pointsize() const { return p_pointsize; };
  double wordpad() const { return p_wordpad; };
  double charpad() const { return p_charpad; };
  bool kern() const { return p_kern; };
  pdf::vAlign valign() const { return p_valign; };
  pdf::hAlign halign() const { return p_halign; };

  double width() { if(!sizecalculated) calculatesize(); return p_width; };
  double height() { if(!sizecalculated) calculatesize(); return p_height; };
  double boxwidth() const { return p_boxwidth; };
  void wrap();
  void wrap(double nw) {setBoxwidth(nw); wrap();};
  void write(pdfPage *page, double x, double y) const;
  pdfText *vsplit(double heightlimit);
  void dumpsize() const {cerr << " pdfText width=" << p_width << " height=" << p_height << "\n";};
private:
  pdf::hAlign p_halign;
  void calculatesize();
  QCString p_font;
  bool p_italic;
  bool p_bold;
  bool p_underline;
  double p_pointsize;
  double p_wordpad;
  double p_charpad;
  bool p_kern;
  pdf::vAlign p_valign;

  QValueList<pdfTextChunk> chunk;
  pdfdocument *p_document;
  double p_boxwidth;
  bool sizecalculated;
  double p_width;
  double p_height;
};

#endif // PDFTEXT_H
