// -*-C++-*-
// pdfdocument.cc
// $Id: pdfdocument.cc,v 1.1 2004/06/13 00:26:57 steve Exp steve $
//
// Copyright (C) 2001 Steve Atkins <steve@blighty.com>
// Last change: Time-stamp: <25-Jul-2004 20:11:35 steve>
//
//

#include "pdf.h"
#include "pdfdocument.h"
#include "pdfchunk.h"
#include "pdfPages.h"
#include "pdfPage.h"
#include "pdfFont.h"
#include "pdfCatalog.h"
#include "pdfImage.h"
#include "pdfText.h"
#include "pdfGrid.h"

pdfdocument::pdfdocument() : idx(0), prefix("%PDF-1.3\n%\377\377\377\377\n"), currentPage(0), p_pagewidth(612), p_pageheight(792)
{
  catalog = new pdfCatalog(this);
  pages = new pdfPages(this);
  catalog->setPages(pages);
  setMargin(36); // 0.5"
  //  cerr << "Creating document - bottommargin=" << bottomMargin() << "\n";
}

pdfdocument::~pdfdocument()
{
  for(QValueList<pdfchunk*>::const_iterator cit=chunks.begin(); cit!=chunks.end(); ++cit) {
    delete *cit;
  }
}

pdfFont *pdfdocument::useFont(QCString basename)
{
  QMap<QCString,pdfFont*>::iterator cit = fontmap.find(basename);
  if(fontmap.end() != cit)
    return cit.data();

  pdfFont *font = new pdfFont(this, basename);
  fontmap[basename] = font;
  return font;
}

static void stdioShow(QByteArray b, void *user)
{
  FILE *f = (FILE*)user;
  fwrite(b.data(), 1,b.size(), f);
}

void pdfdocument::render(FILE *f)
{
  render(&stdioShow, (void*)f);
}

void pdfdocument::render(void (*show)(QByteArray b, void *user), void *user)
{
  // First, create fonts

  // Second, assign indexes
  QValueList<pdfchunk*> orderedlist;
  // If linearising, we can do that here...
  for(QValueList<pdfchunk*>::const_iterator cit=chunks.begin(); cit!=chunks.end(); ++cit) {
    if((*cit)->index() < 0) {
      (*cit)->setIndex(nextindex());
    }
    orderedlist.append(*cit);
  }

  // Third, render everything

  int ptr;
  QMap<int, int> offset;
  
  {
    QByteArray pref(prefix);
    pref.resize(prefix.length());
    show(pref, user);
    ptr=prefix.length();
  }

  QByteArray endobj(100);
  endobj.resize(sprintf(endobj.data(), "endobj\n"));

  for(QValueList<pdfchunk*>::const_iterator cit=orderedlist.begin(); cit!=orderedlist.end(); ++cit) {
    int i = (*cit)->index();
    offset[i] = ptr;
    QByteArray startobj(100);
    startobj.resize(sprintf(startobj.data(), "%d 0 obj\n", i));
    ptr += startobj.size();
    show(startobj, user);
    //    cerr << "calling render() from pdfdocument::render()\n";
    QByteArray b = (*cit)->render();
    ptr += b.size();
    show(b, user);
    ptr += endobj.size();
    show(endobj, user);
  }
  int startxref = ptr;
  {
    // Store 1..idx
    QByteArray xref(20 * idx + 100);
    char *p = xref.data();
    p += sprintf(p, "xref\n0 %d\n0000000000 65535 f \n", idx+1);
    for(int i=1; i<=idx; ++i) {
      QMap<int, int>::const_iterator cit=offset.find(i);
      if(cit == offset.end()) {
	p += sprintf(p, "0000000000 65535 f \n");
      } else {
	p += sprintf(p, "%010d 00000 n \n", cit.data());
      }
    }
    xref.resize(p - xref.data());
    show(xref, user);
  }
  
  QByteArray trailer(500);
  char *p = trailer.data();
  p+= sprintf(p, "trailer\n<< /Size %d\n", idx+1);
  p+= sprintf(p, "/Root %d 0 R\n", catalog->index());
  p+= sprintf(p, ">>\nstartxref\n%d\n%%EOF\n", startxref);
  trailer.resize(p-trailer.data());
  show(trailer, user);
}

int pdfdocument::getFontRef(QCString basename)
{
  QMap<QCString,pdfFont*>::const_iterator cit = fontmap.find(basename);
  if(cit!=fontmap.end()) {
    return cit.data()->index();
  }
  THROW("Non configured font " + basename + " used");
  return 0;
}

pdfPage *pdfdocument::newPage()
{
  pdfPage *page = new pdfPage(this, p_pagewidth, p_pageheight);
  pages->insertPage(page);
  x = p_leftmargin;
  y = p_pageheight - p_topmargin;
  return page;
}

void pdfdocument::append(QCString uid, const QImage& image)
{
  //  cerr << "pdfdocument::append(" << uid << ")\n";
  if(existsXobject(uid))
    return;

  pdfImage *im = new pdfImage(this, image);
  xobjects[uid] = im;
  // append(im); - already done by constructor
}

double pdfdocument::alignx(pdf::hAlign align, double width)
{
  return leftMargin();
  double x = 0.0;
  switch(align) {
  case pdf::hJustify:
  case pdf::hLeft:
    //    cerr << "          alignx Justify or Left\n";
    x = leftMargin();
    break;
  case pdf::hRight:
    //    cerr << "          alignx right\n";
    x = currentPage->width() - rightMargin() - width;
    break;
  case pdf::hCentre:
    //    cerr << "          alignx centre\n";
    x = (currentPage->width() + rightMargin() - leftMargin() - width) / 2;
    break;
  }
  //  cerr << "                   = " << x << "\n";
  return x;
}

void pdfdocument::text(pdfText *t, pdf::hAlign align, double width)
{
  //  cerr << "Entering pdfdocument::text()\n";
  if(0==currentPage)
    currentPage = newPage();
  if(width == 0.0)
    width = currentPage->width() - leftMargin() - rightMargin();

  //  cerr << "Wrapping width=" << width << "\n";
  t->wrap(width);
  //  cerr << "Wrapped y=" << y << " bottommargin=" << bottomMargin() << "\n";
  pdfText *remainder = t->vsplit(y-bottomMargin());
  t->write(currentPage, alignx(align, t->width()), y);
  y -= t->height();
  t = remainder;
  while(t) {
    //    cerr << "Loop top\n";
    currentPage = newPage();
    remainder = t->vsplit(y-bottomMargin());
    t->write(currentPage, alignx(align, t->width()), y);
    y -= t->height();
    delete t;
    t = remainder;
  }
  //  cerr << "Leaving pdfdocument::text()\n";
}

void pdfdocument::image(QImage& image, pdf::hAlign align, double width, double height, QCString uid)
{
  if(0==currentPage)
    currentPage = newPage();
  if(width == 0.0 && height != 0.0)
    width = height * image.width() / image.height();
  if(width == 0.0)
    width = currentPage->width() - leftMargin() - rightMargin();  
  if(height == 0.0)
    height = width * image.height() / image.width();
  if(uid == "") {
    uid.setNum(++imgtagspin);
    uid = "img" + uid;
  }
  
  if(y-height < bottomMargin()) {
    if(y <= currentPage->height() - topMargin() - 0.001) {
      currentPage = newPage();
    }
    if(y-height < bottomMargin()) {
      height = currentPage->height() - topMargin() - bottomMargin();
      width = height * image.width() / image.height();
    }
  }

  currentPage->insertImage(uid, image);
  currentPage->baseContent()->image(uid, alignx(align, width), y, width, height);

  y -= height;
}

void pdfdocument::grid(pdfGrid *g, pdf::hAlign align, double width)
{
  if(0==currentPage)
    currentPage = newPage();
  if(width == 0.0)
    width = currentPage->width() - leftMargin() - rightMargin();

  g->wrap(width);

  int startrow=0;
  int endrow=g->slice(y - bottomMargin(), startrow);
  y -= g->writeSlice(currentPage, startrow, endrow, alignx(align, width), y);
  
  startrow=endrow;
  while(startrow < g->rows()) {
    currentPage = newPage();

    endrow=g->slice(y - bottomMargin(), startrow);
    y -= g->writeSlice(currentPage, startrow, endrow, alignx(align, width), y);
  
    startrow=endrow;
  }
}
