// -*-C++-*-
// pdfdocument.h
// $Id: pdfdocument.h,v 1.1 2004/06/13 00:26:57 steve Exp steve $
//
// Copyright (C) 2001 Steve Atkins <steve@blighty.com>
// Last change: Time-stamp: <11-Jul-2004 18:01:17 steve>
//
//

#ifndef PDFDOCUMENT_H
#define PDFDOCUMENT_H

#include <qvaluelist.h>
#include <qcstring.h>
#include <stdio.h>
#include <qmap.h>
#include <qstring.h>
#include <qimage.h>

#include "pdf.h"

class pdfPage;
class pdfPages;
class pdfCatalog;
class pdfchunk;
class pdfFont;
class pdfXobject;
class pdfText;
class pdfGrid;

class pdfdocument
{
public:
  pdfdocument();
  virtual ~pdfdocument();

  virtual const char * isA() const { return "pdfdocument"; };
  int nextindex() { return ++idx; };
  QValueList<QByteArray> render();
  void render(FILE *f);
  void render(void (*show)(QByteArray b, void *user), void *user);

  void append(pdfchunk *kid) {chunks.append(kid); };
  
  int getFontRef(QCString basename);
  pdfFont *useFont(QCString basefont);
  pdfPage *newPage();
  bool existsXobject(QCString s) const {return xobjects.contains(s);};
  pdfXobject *getXobject(QCString uid) const {return xobjects[uid];};
  void append(QCString uid, const QImage& image);

  void text(pdfText *t, pdf::hAlign align=pdf::hLeft, double width=0.0);
  void image(QImage& image, pdf::hAlign align=pdf::hLeft, double width=0.0, double height=0.0, QCString uid="");
  void grid(pdfGrid *g, pdf::hAlign align=pdf::hLeft, double width=0.0);

  void setLeftMargin(double b) {p_leftmargin = b;};
  void setRightMargin(double b) {p_rightmargin = b;};
  void setTopMargin(double b) {p_topmargin = b;};
  void setBottomMargin(double b) {p_bottommargin = b;};
  void setMargin(double b) {setLeftMargin(b);setRightMargin(b);setTopMargin(b);setBottomMargin(b);};

  double leftMargin() const {return p_leftmargin;};
  double rightMargin() const {return p_rightmargin;};
  double topMargin() const {return p_topmargin;};
  double bottomMargin() const {return p_bottommargin;};
  double width() const {return p_pagewidth;};
  double usableWidth() const {return width()-leftMargin()-rightMargin();};

private:
  double alignx(pdf::hAlign align, double width);
  QValueList<pdfchunk*> chunks;
  int idx;
  QCString prefix;
  pdfCatalog *catalog;
  pdfPages *pages;

  pdfPage *currentPage;
  double p_pagewidth;
  double p_pageheight;
  double p_leftmargin;
  double p_rightmargin;
  double p_topmargin;
  double p_bottommargin;

  double x;
  double y;

  int imgtagspin;

  QMap<QCString, pdfFont*> fontmap;
  QMap<QCString, pdfXobject*> xobjects;
};


#endif // PDFDOCUMENT_H
